from typing import Union
from ..common import Expr, InstrumentParameter
from ..instr import Instr, Instance
from ..reader import Reader, Registry
from loguru import logger
from mccode_antlr.instr.orientation import Vector, Angles

class Assembler:
    """Interactive instrument assembly"""

    def __init__(self, name: str, registries: list[Registry] = None):
        from ..reader.registry import ordered_registries
        if registries is not None:
            registries = list(ordered_registries(registries))
        self.instrument = Instr(name, source='interactive')
        self.reader = Reader(registries=registries) if registries is not None else Reader()
        self.instrument.registries = self.reader.registries

    @property
    def name(self):
        return self.instrument.name

    def _handle_at_rotate(self, a=None) -> tuple[tuple[Expr, Expr, Expr], Union[Instance,  None]]:
        if a is None:
            return (Expr.float(0), Expr.float(0), Expr.float(0)), None
        if hasattr(a, '__len__') and len(a) == 3:
            a = a, 'ABSOLUTE'
        if not hasattr(a, '__len__') or len(a) != 2:
            raise RuntimeError('Require two-tuple of three values and a reference')
        v, ref = a
        if ref is not None and not isinstance(ref, Instance):
            if 'absolute' == ref.lower():
                ref = None
            elif isinstance(ref, str):
                # Get the component instance by name
                ref = self.instrument.get_component(ref)
            else:
                raise RuntimeError(f'No logic pathway for instance reference {ref}')
        if not hasattr(v, '__len__') or len(v) != 3:
            raise RuntimeError('Position/orientation must have three elements')
        v = tuple(self.instrument.check_expr(x) for x in v)
        return (v[0], v[1], v[2]), ref

    def _handle_at(self, a=None) -> tuple[Vector, Union[Instance, None]]:
        at_tuple, ref = self._handle_at_rotate(a)
        return Vector(*at_tuple), ref

    def _handle_rotate(self, a=None, at_ref=None) -> tuple[Angles, Union[Instance, None]]:
        rot_tuple, ref = self._handle_at_rotate(a)
        return Angles(*rot_tuple), ref or at_ref

    def component(self, name: str, type_name: str, at=None, rotate=None, parameters=None, **kwargs):
        """Add a component to the underlying Instr.

        Parameters
        ----------
        name : str
            The name of the component instance.
        type_name : str
            The name of the component type.
        at : tuple, optional
            The position and orientation of the component instance.
            If not provided, the component will be placed at the origin.
        rotate : tuple, optional
            The rotation of the component instance.
            If not provided, the component will be rotated to match the at argument.
        parameters : dict, optional
            A dictionary of parameter names and values to set for the component instance.
        kwargs : dict, optional
            Properties for the constructed component Instance object. Useful keyword values include
            `when` and `group`.

        Note
        ----
        See `Assembler._handle_at` and `Assembler._handle_rotate` for details on the at and rotate arguments.
        """
        comp_type = self.reader.get_component(type_name)
        if type_name != comp_type.name:
            raise RuntimeError(f"Component resolution failed for {type_name}, found {comp_type.name} instead")
        if self.reader.c_flags:
            unique_flags = list(self.instrument.flags)
            unique_flags.extend(self.reader.c_flags)
            self.instrument.flags = tuple(dict.fromkeys(unique_flags))
        at, ref = self._handle_at(at)
        instance = Instance(name, comp_type,
                            at_relative=(at, ref), rotate_relative=self._handle_rotate(rotate, ref),
                            **kwargs)
        self.instrument.add_component(instance)
        if isinstance(parameters, dict):
            instance.set_parameters(**parameters)
        return instance

    def parameter(self, par, ignore_repeated=False):
        """Add a parameter to the underlying Instr.

        Note
        ----
        The ignore_repeated keyword argument can be set to True in order to merely ensure that a parameter exists.
        Otherwise, repeatedly specifying the same parameter will raise a RuntimeError.
        """
        if isinstance(par, str):
            par = InstrumentParameter.parse(par)
        if not isinstance(par, InstrumentParameter):
            logger.warning(f'Unhandled parameter {par}')
        self.instrument.add_parameter(par, ignore_repeated=ignore_repeated)

    def parameters(self, *pars, **pairs):
        for par in list(pars):
            if isinstance(par, str):
                par = InstrumentParameter.parse(par)
            if not isinstance(par, InstrumentParameter):
                logger.warning(f'Unhandled parameter(s) {par}')
            self.instrument.add_parameter(par)

        for name, value in pairs.items():
            if not isinstance(value, InstrumentParameter):
                if isinstance(value, dict) and 'unit' in value and 'value' in value:
                    value, unit = value['value'], value['unit']
                elif isinstance(value, (list, tuple)) and len(value) == 2 and isinstance(value[1], str):
                    value, unit = value
                else:
                    unit = ''
                value = InstrumentParameter(name, unit, value if isinstance(value, Expr) else Expr.best(value))
            self.instrument.add_parameter(value)

    def declare(self, string, source=None, line=-1):
        return _rawc_call(self.instrument.DECLARE, string, source, line)

    def declare_array(self, dtype: str, name: str, init: list, source=None, line=-1):
        return self.declare(f'{dtype} {name}[] = {{{",".join(str(x) for x in init)}}};', source=source, line=line)

    def user_vars(self, string, source=None, line=-1):
        return _rawc_call(self.instrument.USERVARS, string, source, line)

    def ensure_user_var(self, string, source=None, line=-1):
        # tying the Assembler to work with C might not be great
        from mccode_antlr.translators.c_listener import extract_c_declared_variables as parse
        variables = parse(string)
        if len(variables) == 0:
            raise ValueError(f'The provided input {string} does not specify a C parameter declaration.')
        if len(variables) != 1:
            print(f'The provided input {string} specifies {len(variables)} C parameter declarations, using only the first')
        decl = variables[0]
        name = decl.name
        dtype = decl.dtype
        for user_vars in self.instrument.user:
            uv_variables = parse(user_vars.source)
            if any(x.dtype == dtype and x.name == name for x in uv_variables):
                return
            if any(x.name == name for x in uv_variables):
                print(f'A USERVARS variable with name {name} but type different than {dtype} has already been defined.')
                return
        return self.user_vars(string, source=source, line=line)

    def initialize(self, string, source=None, line=-1):
        return _rawc_call(self.instrument.INITIALIZE, string, source, line)

    def save(self, string, source=None, line=-1):
        return _rawc_call(self.instrument.SAVE, string, source, line)

    def final(self, string, source=None, line=-1):
        return _rawc_call(self.instrument.FINALLY, string, source, line)

    def metadata(self, name: str, mimetype: str, value: str, source=None):
        from mccode_antlr.common.metadata import MetaData
        self.instrument.add_metadata(MetaData.from_instrument_tokens(source, mimetype, name, value))


def _rawc_call(method, string: str, source: str = None, line: int = -1):
    from mccode_antlr.common import RawC
    return method(RawC(str(source), line, string))


INTENDED_USAGE = """
bifrost = Assembler('bifrost', registries=[MCSTAS_REGISTRY, local_bifrost_components])
bifrost.parameters(par1=5.11, par2=('m', 100), par3={'value': 3.14159, 'unit': 'radian'})
...
bifrost.component('source, 'ESS_BUTTERFLY').set_parameters(...)
...
"""


# Generated from /home/g/PycharmProjects/mccode-antlr/src/grammar/McComp.g4 by ANTLR 4.13.2
from antlr4 import *
if "." in __name__:
    from .McCompParser import McCompParser
else:
    from McCompParser import McCompParser

# This class defines a complete generic visitor for a parse tree produced by McCompParser.

class McCompVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by McCompParser#prog.
    def visitProg(self, ctx:McCompParser.ProgContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#ComponentDefineNew.
    def visitComponentDefineNew(self, ctx:McCompParser.ComponentDefineNewContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#ComponentDefineCopy.
    def visitComponentDefineCopy(self, ctx:McCompParser.ComponentDefineCopyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#TraceBlock.
    def visitTraceBlock(self, ctx:McCompParser.TraceBlockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#TraceBlockCopy.
    def visitTraceBlockCopy(self, ctx:McCompParser.TraceBlockCopyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#component_parameter_set.
    def visitComponent_parameter_set(self, ctx:McCompParser.Component_parameter_setContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#component_define_parameters.
    def visitComponent_define_parameters(self, ctx:McCompParser.Component_define_parametersContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#component_set_parameters.
    def visitComponent_set_parameters(self, ctx:McCompParser.Component_set_parametersContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#component_out_parameters.
    def visitComponent_out_parameters(self, ctx:McCompParser.Component_out_parametersContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#component_parameters.
    def visitComponent_parameters(self, ctx:McCompParser.Component_parametersContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#ComponentParameterDouble.
    def visitComponentParameterDouble(self, ctx:McCompParser.ComponentParameterDoubleContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#ComponentParameterInteger.
    def visitComponentParameterInteger(self, ctx:McCompParser.ComponentParameterIntegerContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#ComponentParameterString.
    def visitComponentParameterString(self, ctx:McCompParser.ComponentParameterStringContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#ComponentParameterVector.
    def visitComponentParameterVector(self, ctx:McCompParser.ComponentParameterVectorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#ComponentParameterSymbol.
    def visitComponentParameterSymbol(self, ctx:McCompParser.ComponentParameterSymbolContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#ComponentParameterDoubleArray.
    def visitComponentParameterDoubleArray(self, ctx:McCompParser.ComponentParameterDoubleArrayContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#ComponentParameterIntegerArray.
    def visitComponentParameterIntegerArray(self, ctx:McCompParser.ComponentParameterIntegerArrayContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#ShareBlock.
    def visitShareBlock(self, ctx:McCompParser.ShareBlockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#ShareBlockCopy.
    def visitShareBlockCopy(self, ctx:McCompParser.ShareBlockCopyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#DisplayBlock.
    def visitDisplayBlock(self, ctx:McCompParser.DisplayBlockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#DisplayBlockCopy.
    def visitDisplayBlockCopy(self, ctx:McCompParser.DisplayBlockCopyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#component_ref.
    def visitComponent_ref(self, ctx:McCompParser.Component_refContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#coords.
    def visitCoords(self, ctx:McCompParser.CoordsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#reference.
    def visitReference(self, ctx:McCompParser.ReferenceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#dependency.
    def visitDependency(self, ctx:McCompParser.DependencyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#DeclareBlock.
    def visitDeclareBlock(self, ctx:McCompParser.DeclareBlockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#DeclareBlockCopy.
    def visitDeclareBlockCopy(self, ctx:McCompParser.DeclareBlockCopyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#uservars.
    def visitUservars(self, ctx:McCompParser.UservarsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#InitializeBlock.
    def visitInitializeBlock(self, ctx:McCompParser.InitializeBlockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#InitializeBlockCopy.
    def visitInitializeBlockCopy(self, ctx:McCompParser.InitializeBlockCopyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#SaveBlock.
    def visitSaveBlock(self, ctx:McCompParser.SaveBlockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#SaveBlockCopy.
    def visitSaveBlockCopy(self, ctx:McCompParser.SaveBlockCopyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#FinallyBlock.
    def visitFinallyBlock(self, ctx:McCompParser.FinallyBlockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#FinallyBlockCopy.
    def visitFinallyBlockCopy(self, ctx:McCompParser.FinallyBlockCopyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#metadata.
    def visitMetadata(self, ctx:McCompParser.MetadataContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#category.
    def visitCategory(self, ctx:McCompParser.CategoryContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#initializerlist.
    def visitInitializerlist(self, ctx:McCompParser.InitializerlistContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#assignment.
    def visitAssignment(self, ctx:McCompParser.AssignmentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#ExpressionBinaryMod.
    def visitExpressionBinaryMod(self, ctx:McCompParser.ExpressionBinaryModContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#ExpressionBinaryLess.
    def visitExpressionBinaryLess(self, ctx:McCompParser.ExpressionBinaryLessContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#ExpressionBinaryGreater.
    def visitExpressionBinaryGreater(self, ctx:McCompParser.ExpressionBinaryGreaterContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#ExpressionBinaryLessEqual.
    def visitExpressionBinaryLessEqual(self, ctx:McCompParser.ExpressionBinaryLessEqualContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#ExpressionArrayAccess.
    def visitExpressionArrayAccess(self, ctx:McCompParser.ExpressionArrayAccessContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#ExpressionBinaryLogic.
    def visitExpressionBinaryLogic(self, ctx:McCompParser.ExpressionBinaryLogicContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#ExpressionInteger.
    def visitExpressionInteger(self, ctx:McCompParser.ExpressionIntegerContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#ExpressionBinaryRightShift.
    def visitExpressionBinaryRightShift(self, ctx:McCompParser.ExpressionBinaryRightShiftContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#ExpressionMyself.
    def visitExpressionMyself(self, ctx:McCompParser.ExpressionMyselfContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#ExpressionPrevious.
    def visitExpressionPrevious(self, ctx:McCompParser.ExpressionPreviousContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#ExpressionIdentifier.
    def visitExpressionIdentifier(self, ctx:McCompParser.ExpressionIdentifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#ExpressionStructAccess.
    def visitExpressionStructAccess(self, ctx:McCompParser.ExpressionStructAccessContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#ExpressionFunctionCall.
    def visitExpressionFunctionCall(self, ctx:McCompParser.ExpressionFunctionCallContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#ExpressionBinaryMD.
    def visitExpressionBinaryMD(self, ctx:McCompParser.ExpressionBinaryMDContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#ExpressionString.
    def visitExpressionString(self, ctx:McCompParser.ExpressionStringContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#ExpressionGrouping.
    def visitExpressionGrouping(self, ctx:McCompParser.ExpressionGroupingContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#ExpressionExponentiation.
    def visitExpressionExponentiation(self, ctx:McCompParser.ExpressionExponentiationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#ExpressionBinaryLeftShift.
    def visitExpressionBinaryLeftShift(self, ctx:McCompParser.ExpressionBinaryLeftShiftContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#ExpressionBinaryGreaterEqual.
    def visitExpressionBinaryGreaterEqual(self, ctx:McCompParser.ExpressionBinaryGreaterEqualContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#ExpressionZero.
    def visitExpressionZero(self, ctx:McCompParser.ExpressionZeroContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#ExpressionUnaryPM.
    def visitExpressionUnaryPM(self, ctx:McCompParser.ExpressionUnaryPMContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#ExpressionTrinaryLogic.
    def visitExpressionTrinaryLogic(self, ctx:McCompParser.ExpressionTrinaryLogicContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#ExpressionFloat.
    def visitExpressionFloat(self, ctx:McCompParser.ExpressionFloatContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#ExpressionPointerAccess.
    def visitExpressionPointerAccess(self, ctx:McCompParser.ExpressionPointerAccessContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#ExpressionBinaryEqual.
    def visitExpressionBinaryEqual(self, ctx:McCompParser.ExpressionBinaryEqualContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#ExpressionBinaryPM.
    def visitExpressionBinaryPM(self, ctx:McCompParser.ExpressionBinaryPMContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#ExpressionUnaryLogic.
    def visitExpressionUnaryLogic(self, ctx:McCompParser.ExpressionUnaryLogicContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#shell.
    def visitShell(self, ctx:McCompParser.ShellContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#SearchPath.
    def visitSearchPath(self, ctx:McCompParser.SearchPathContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#SearchShell.
    def visitSearchShell(self, ctx:McCompParser.SearchShellContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McCompParser#unparsed_block.
    def visitUnparsed_block(self, ctx:McCompParser.Unparsed_blockContext):
        return self.visitChildren(ctx)



del McCompParser


from mccode_antlr.reader import Registry


def mccode_script_parse(prog: str):
    from argparse import ArgumentParser
    from pathlib import Path

    def resolvable(name: str):
        return None if name is None else Path(name).resolve()

    parser = ArgumentParser(prog=prog, description=f'Convert mccode_antlr-3 instr and comp files to {prog} runtime in C')
    parser.add_argument('filename', type=resolvable, nargs='?', help='.instr file name to be converted')

    parser.add_argument('-o', '--output-file', type=str, help='Output filename for C runtime file')
    parser.add_argument('-I', '--search-dir', action='append', type=resolvable, help='Extra component search directory')
    parser.add_argument('-t', '--trace', action='store_true', help="Enable 'trace' mode for instrument display")
    parser.add_argument('-p', '--portable', action='store_true', help='No idea. Your guess is better than mine.')
    parser.add_argument('-v', '--version', action='store_true', help='Print the McCode version')
    parser.add_argument('--source', action='store_true', help='Embed the instrument source code in the executable')
    parser.add_argument('--no-main', action='store_true', help='Do not create main(), for external embedding')
    parser.add_argument('--no-runtime', action='store_true', help='Do not embed run-time libraries')
    parser.add_argument('--verbose', action='store_true', help='Verbose output during conversion')

    args = parser.parse_args()

    if args.version:
        from sys import exit
        from mccode_antlr.version import version
        print(f'mccode_antlr code generator version {version()}')
        print(' Copyright (c) European Spallation Source ERIC, 2023-2024')
        print('Based on McStas/McXtrace version 3')
        print(' Copyright (c) DTU Physics and Risoe National Laboratory, 1997-2023')
        print(' Additions (c) Institut Laue Langevin, 2003-2019')
        print('All rights reserved\n\nComponents are (c) their authors, see component headers.')
        exit(1)

    if args.filename is None:
        parser.error('No input file provided')

    return args


def mccode(flavor: str, registry: Registry, generator: dict):
    from pathlib import Path
    from mccode_antlr.reader import Reader
    from mccode_antlr.reader import LocalRegistry
    from mccode_antlr.translators.c import CTargetVisitor
    from mccode_antlr.common import Mode

    args = mccode_script_parse(flavor)

    config = dict(default_main=(not args.no_main) if args.no_main is not None else True,
                  enable_trace=args.trace if args.trace is not None else False,
                  portable=args.portable if args.portable is not None else False,
                  include_runtime=(not args.no_runtime) if args.no_runtime is not None else True,
                  embed_instrument_file=args.source if args.source is not None else False,
                  verbose=args.verbose if args.verbose is not None else False,
                  output=args.output_file if args.output_file is not None else args.filename.with_suffix('.c')
                  )
    # McCode always requires access to a remote Pooch repository:
    registries = [registry]
    # A user can specify extra (local) directories to search for included files using -I or --search-dir
    if args.search_dir is not None and len(args.search_dir):
        registries.extend([LocalRegistry(d.stem, d) for d in args.search_dir])
    # And McCode-3 users expect to always have access to files in the current working directory
    registries.append(LocalRegistry('working_directory', f'{Path().resolve()}'))

    # Construct the object which will read the instrument and component files, producing Python objects
    reader = Reader(registries=registries)
    # Read the provided .instr file, including all specified .instr and .comp files along the way
    # In minimal mode, the component orientations are not resolved -- to speed up the process
    instrument = reader.get_instrument(args.filename, mode=Mode.minimal)
    # Construct the object which will translate the Python instrument to C
    visitor = CTargetVisitor(instrument, generate=generator, config=config, verbose=config['verbose'])
    # Go through the instrument, finish by writing the output file:
    visitor.save(filename=config['output'])


def mcstas():
    from mccode_antlr.reader import MCSTAS_REGISTRY
    from mccode_antlr.translators.target import MCSTAS_GENERATOR
    mccode('mcstas', MCSTAS_REGISTRY, MCSTAS_GENERATOR)


def mcxtrace():
    from mccode_antlr.reader import MCXTRACE_REGISTRY
    from mccode_antlr.translators.target import MCXTRACE_GENERATOR
    mccode('mcxtrace', MCXTRACE_REGISTRY, MCXTRACE_GENERATOR)



def cogen_save(source, declared_parameters):
    lines = ["/* *****************************************************************************",
             f"* instrument '{source.name}' and components SAVE",
             "***************************************************************************** */",
             ]

    for comp in source.component_types():
        lines.extend(cogen_comp_save_class(comp, declared_parameters[comp.name]))

    # write the instrument main code, which calls component ones
    lines.extend([
        f'int save(FILE *handle) {{ /* called by mccode_main for {source.name}:SAVE */',
        "  if (!handle) siminfo_init(NULL);",
    ])

    # insert user code from instrument definition
    if len(source.save):
        f, n = source.save[0].fn
        lines.extend([
            f'  /* Instrument {source.name} SAVE */',
            f'  SIG_MESSAGE("[{source.name} SAVE [{f}:{n}]");'
        ])
        for par in source.parameters:
            # ensure there's no conflict of names
            lines.append(f'  #define {par.name} (instrument->_parameters.{par.name}')
        for block in source.save:
            lines.append(block.to_c())
        for par in source.parameters:
            lines.append(f'  #undef {par.name}')

    lines.append('  /* call iteratively all components SAVE */')
    for comp in source.components:
        if len(comp.type.save):
            lines.append(f'  class_{comp.type.name}_save(&_{comp.name}_var);')

    lines.extend([
        '  if (!handle) siminfo_close();'
        '',
        '  return(0);',
        '} /* save */',
        ''
    ])
    return '\n'.join(lines)


def cogen_comp_save_class(comp, declared_parameters):
    from .c_defines import cogen_parameter_define, cogen_parameter_undef
    if not len(comp.save):
        return []

    lines = [
        # f'_class_{comp.name} *class_{comp.name}_save(_class_{comp.name} *_comp) {{',
        # Changed following https://github.com/McStasMcXtrace/McCode/commit/73a32858af1a5193bbb692da49379816cb0f9e38
        f'void class_{comp.name}_save(_class_{comp.name} *_comp) {{',
        cogen_parameter_define(comp, declared_parameters)
    ]
    f, n = comp.initialize[0].fn if len(comp.initialize) else (comp.name, 0)
    lines.append(f'  SIG_MESSAGE("[_{comp.name}_save] component NULL={comp.name}() [{f}:{n}]");')

    for block in comp.save:
        lines.append(block.to_c())

    lines.extend([
        cogen_parameter_undef(comp, declared_parameters),
        # '  return(_comp);',
        '  return;',
        f'}} /* class_{comp.name}_save */',
        ''
    ])
    return lines



# Generated from /home/g/PycharmProjects/mccode-antlr/src/grammar/McInstr.g4 by ANTLR 4.13.2
from antlr4 import *
if "." in __name__:
    from .McInstrParser import McInstrParser
else:
    from McInstrParser import McInstrParser

# This class defines a complete generic visitor for a parse tree produced by McInstrParser.

class McInstrVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by McInstrParser#prog.
    def visitProg(self, ctx:McInstrParser.ProgContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#instrument_definition.
    def visitInstrument_definition(self, ctx:McInstrParser.Instrument_definitionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#instrument_parameters.
    def visitInstrument_parameters(self, ctx:McInstrParser.Instrument_parametersContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#InstrumentParameterDouble.
    def visitInstrumentParameterDouble(self, ctx:McInstrParser.InstrumentParameterDoubleContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#InstrumentParameterInteger.
    def visitInstrumentParameterInteger(self, ctx:McInstrParser.InstrumentParameterIntegerContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#InstrumentParameterString.
    def visitInstrumentParameterString(self, ctx:McInstrParser.InstrumentParameterStringContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#instrument_parameter_unit.
    def visitInstrument_parameter_unit(self, ctx:McInstrParser.Instrument_parameter_unitContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#instrument_trace.
    def visitInstrument_trace(self, ctx:McInstrParser.Instrument_traceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#instrument_metadata.
    def visitInstrument_metadata(self, ctx:McInstrParser.Instrument_metadataContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#instrument_trace_include.
    def visitInstrument_trace_include(self, ctx:McInstrParser.Instrument_trace_includeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#component_instance.
    def visitComponent_instance(self, ctx:McInstrParser.Component_instanceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#InstanceNameCopyIdentifier.
    def visitInstanceNameCopyIdentifier(self, ctx:McInstrParser.InstanceNameCopyIdentifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#InstanceNameCopy.
    def visitInstanceNameCopy(self, ctx:McInstrParser.InstanceNameCopyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#InstanceNameIdentifier.
    def visitInstanceNameIdentifier(self, ctx:McInstrParser.InstanceNameIdentifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#ComponentTypeCopy.
    def visitComponentTypeCopy(self, ctx:McInstrParser.ComponentTypeCopyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#ComponentTypeIdentifier.
    def visitComponentTypeIdentifier(self, ctx:McInstrParser.ComponentTypeIdentifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#instance_parameters.
    def visitInstance_parameters(self, ctx:McInstrParser.Instance_parametersContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#InstanceParameterExpr.
    def visitInstanceParameterExpr(self, ctx:McInstrParser.InstanceParameterExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#InstanceParameterNull.
    def visitInstanceParameterNull(self, ctx:McInstrParser.InstanceParameterNullContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#InstanceParameterVector.
    def visitInstanceParameterVector(self, ctx:McInstrParser.InstanceParameterVectorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#split.
    def visitSplit(self, ctx:McInstrParser.SplitContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#when.
    def visitWhen(self, ctx:McInstrParser.WhenContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#place.
    def visitPlace(self, ctx:McInstrParser.PlaceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#orientation.
    def visitOrientation(self, ctx:McInstrParser.OrientationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#groupref.
    def visitGroupref(self, ctx:McInstrParser.GrouprefContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#jumps.
    def visitJumps(self, ctx:McInstrParser.JumpsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#jump.
    def visitJump(self, ctx:McInstrParser.JumpContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#JumpPrevious.
    def visitJumpPrevious(self, ctx:McInstrParser.JumpPreviousContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#JumpMyself.
    def visitJumpMyself(self, ctx:McInstrParser.JumpMyselfContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#JumpNext.
    def visitJumpNext(self, ctx:McInstrParser.JumpNextContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#JumpIdentifier.
    def visitJumpIdentifier(self, ctx:McInstrParser.JumpIdentifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#extend.
    def visitExtend(self, ctx:McInstrParser.ExtendContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#component_ref.
    def visitComponent_ref(self, ctx:McInstrParser.Component_refContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#coords.
    def visitCoords(self, ctx:McInstrParser.CoordsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#reference.
    def visitReference(self, ctx:McInstrParser.ReferenceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#dependency.
    def visitDependency(self, ctx:McInstrParser.DependencyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#DeclareBlock.
    def visitDeclareBlock(self, ctx:McInstrParser.DeclareBlockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#DeclareBlockCopy.
    def visitDeclareBlockCopy(self, ctx:McInstrParser.DeclareBlockCopyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#uservars.
    def visitUservars(self, ctx:McInstrParser.UservarsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#InitializeBlock.
    def visitInitializeBlock(self, ctx:McInstrParser.InitializeBlockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#InitializeBlockCopy.
    def visitInitializeBlockCopy(self, ctx:McInstrParser.InitializeBlockCopyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#SaveBlock.
    def visitSaveBlock(self, ctx:McInstrParser.SaveBlockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#SaveBlockCopy.
    def visitSaveBlockCopy(self, ctx:McInstrParser.SaveBlockCopyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#FinallyBlock.
    def visitFinallyBlock(self, ctx:McInstrParser.FinallyBlockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#FinallyBlockCopy.
    def visitFinallyBlockCopy(self, ctx:McInstrParser.FinallyBlockCopyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#metadata.
    def visitMetadata(self, ctx:McInstrParser.MetadataContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#category.
    def visitCategory(self, ctx:McInstrParser.CategoryContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#initializerlist.
    def visitInitializerlist(self, ctx:McInstrParser.InitializerlistContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#assignment.
    def visitAssignment(self, ctx:McInstrParser.AssignmentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#ExpressionBinaryMod.
    def visitExpressionBinaryMod(self, ctx:McInstrParser.ExpressionBinaryModContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#ExpressionBinaryLess.
    def visitExpressionBinaryLess(self, ctx:McInstrParser.ExpressionBinaryLessContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#ExpressionBinaryGreater.
    def visitExpressionBinaryGreater(self, ctx:McInstrParser.ExpressionBinaryGreaterContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#ExpressionBinaryLessEqual.
    def visitExpressionBinaryLessEqual(self, ctx:McInstrParser.ExpressionBinaryLessEqualContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#ExpressionArrayAccess.
    def visitExpressionArrayAccess(self, ctx:McInstrParser.ExpressionArrayAccessContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#ExpressionBinaryLogic.
    def visitExpressionBinaryLogic(self, ctx:McInstrParser.ExpressionBinaryLogicContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#ExpressionInteger.
    def visitExpressionInteger(self, ctx:McInstrParser.ExpressionIntegerContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#ExpressionBinaryRightShift.
    def visitExpressionBinaryRightShift(self, ctx:McInstrParser.ExpressionBinaryRightShiftContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#ExpressionMyself.
    def visitExpressionMyself(self, ctx:McInstrParser.ExpressionMyselfContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#ExpressionPrevious.
    def visitExpressionPrevious(self, ctx:McInstrParser.ExpressionPreviousContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#ExpressionIdentifier.
    def visitExpressionIdentifier(self, ctx:McInstrParser.ExpressionIdentifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#ExpressionStructAccess.
    def visitExpressionStructAccess(self, ctx:McInstrParser.ExpressionStructAccessContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#ExpressionFunctionCall.
    def visitExpressionFunctionCall(self, ctx:McInstrParser.ExpressionFunctionCallContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#ExpressionBinaryMD.
    def visitExpressionBinaryMD(self, ctx:McInstrParser.ExpressionBinaryMDContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#ExpressionString.
    def visitExpressionString(self, ctx:McInstrParser.ExpressionStringContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#ExpressionGrouping.
    def visitExpressionGrouping(self, ctx:McInstrParser.ExpressionGroupingContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#ExpressionExponentiation.
    def visitExpressionExponentiation(self, ctx:McInstrParser.ExpressionExponentiationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#ExpressionBinaryLeftShift.
    def visitExpressionBinaryLeftShift(self, ctx:McInstrParser.ExpressionBinaryLeftShiftContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#ExpressionBinaryGreaterEqual.
    def visitExpressionBinaryGreaterEqual(self, ctx:McInstrParser.ExpressionBinaryGreaterEqualContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#ExpressionZero.
    def visitExpressionZero(self, ctx:McInstrParser.ExpressionZeroContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#ExpressionUnaryPM.
    def visitExpressionUnaryPM(self, ctx:McInstrParser.ExpressionUnaryPMContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#ExpressionTrinaryLogic.
    def visitExpressionTrinaryLogic(self, ctx:McInstrParser.ExpressionTrinaryLogicContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#ExpressionFloat.
    def visitExpressionFloat(self, ctx:McInstrParser.ExpressionFloatContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#ExpressionPointerAccess.
    def visitExpressionPointerAccess(self, ctx:McInstrParser.ExpressionPointerAccessContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#ExpressionBinaryEqual.
    def visitExpressionBinaryEqual(self, ctx:McInstrParser.ExpressionBinaryEqualContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#ExpressionBinaryPM.
    def visitExpressionBinaryPM(self, ctx:McInstrParser.ExpressionBinaryPMContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#ExpressionUnaryLogic.
    def visitExpressionUnaryLogic(self, ctx:McInstrParser.ExpressionUnaryLogicContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#shell.
    def visitShell(self, ctx:McInstrParser.ShellContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#SearchPath.
    def visitSearchPath(self, ctx:McInstrParser.SearchPathContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#SearchShell.
    def visitSearchShell(self, ctx:McInstrParser.SearchShellContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by McInstrParser#unparsed_block.
    def visitUnparsed_block(self, ctx:McInstrParser.Unparsed_blockContext):
        return self.visitChildren(ctx)



del McInstrParser



# def cogen_section(instr: Instr, section: str)

    # write out some macro stuff if this is DISPLAY

    # write out some stuff if this is INITIALISE

    # Iterate through components,
    #  Call the same function with different arguments for each component
    #  exact call differs for INITIALISE, SAVE, FINALLY, DISPLAY

    # write out some undef macro stuff if this is DISPLAY

    # Write out a function signature
    # INITIALISE ->  int initalize(void)
    # SAVE       ->  int save(File * handle)
    # FINALLY    ->  int finally(void)
    # DISPLAY    ->  int display(void)

    # Write out different things for each of INITIALISE, SAVE, FINALLY, DISPLAY

    # If code was provided for this section in the instrument definition, print it now

    # if this is initialise, produce calls to each _{component name}_setpos()

    # Iterate through all components, getting the section code from the component
    # and adding a call to `class_{component type name}_{section}(&_{component name}_var);`
    # if the component instance defined a code block for this section

    # add some section dependent code to the end of the function definition

    # end the function, returning 0 -- always.


import unittest
from textwrap import dedent
from mccode_antlr.loader import parse_mcstas_instr
from .compiled import compiled, gpu_only, compile_and_run


class CRuntimeTestCase(unittest.TestCase):
    def _do_trace_tests(self, with_acc):
        instr = parse_mcstas_instr(dedent(
            """\
            DEFINE INSTRUMENT test_component_traces_visited()
            TRACE
            COMPONENT a = Arm() AT (0, 0, 0) ABSOLUTE EXTEND %{printf("visited a\\n");%}
            COMPONENT b = Arm() AT (0, 0, 1) ABSOLUTE EXTEND %{printf("visited b\\n");%}
            END
            """)
        )
        #
        results, data = compile_and_run(instr, "-n 1", target={'acc': with_acc})
        lines = results.decode('utf-8').splitlines()
        self.assertEqual(1, sum(line == 'visited a' for line in lines))
        self.assertEqual(1, sum(line == 'visited b' for line in lines))

    @compiled
    def test_raytrace_traces_visited(self):
        self._do_trace_tests(with_acc=False)

    @gpu_only
    def test_funnel_raytrace_traces_visited(self):
        # The FUNNEL define is only followed if OpenACC is enabled :/
        # And real OpenACC can only be enabled if we have a GPU :(
        # But since we don't actually need ACC, we can force use of the normal compiler instead
        # Unfortunately the runtime still includes openacc.h, so we should not try to run this normally.
        from mccode_antlr.config import config
        config['acc'] = config['cc'].get(str)
        config['flags']['acc'] = config['flags']['cc'].get(str) + ' -DFUNNEL -DOPENACC -DGCCOFFLOAD'
        self._do_trace_tests(with_acc=True)

    def _do_jump_tests(self, contents: str, jumps: int):
        instr = parse_mcstas_instr(contents)
        results, data = compile_and_run(instr, f"-n 1 jumps={jumps}")
        lines = results.decode('utf-8').splitlines()
        self.assertEqual(jumps, sum(line.startswith('time=') for line in lines))
        times = [int(x.split('=')[-1]) for x in filter(lambda y: y.startswith('time='), lines)]
        self.assertEqual(((jumps + 1) * jumps) >> 1, sum(times))

    @compiled
    def test_jump_iterate(self):
        contents = dedent("""\
                          DEFINE INSTRUMENT test_jump_iterate(int jumps)
                          USERVARS %{int time;%}
                          TRACE
                          COMPONENT o = Arm() AT (0,0,0) ABSOLUTE EXTEND %{time=0;%}
                          COMPONENT a = Arm() AT (0,0,0) ABSOLUTE EXTEND %{time+=1;%}
                          COMPONENT b = Arm() AT (0,0,1) ABSOLUTE EXTEND %{printf("time=%d\\n", time);%}
                          COMPONENT c = Arm() AT (0,0,2) ABSOLUTE JUMP a ITERATE jumps
                          END
                          """)
        self._do_jump_tests(contents, 10)

    @compiled
    def test_jump_when(self):
        contents = dedent("""\
                          DEFINE INSTRUMENT test_jump_when(int jumps)
                          USERVARS %{int time; int do_jump;%}
                          TRACE
                          COMPONENT o = Arm() AT (0,0,0) ABSOLUTE EXTEND %{time=0;%}
                          COMPONENT a = Arm() AT (0,0,0) ABSOLUTE EXTEND %{
                             time+=1;
                             do_jump = time < INSTRUMENT_GETPAR(jumps) ? 1 : 0;
                          %}
                          COMPONENT b = Arm() AT (0,0,1) ABSOLUTE EXTEND %{printf("time=%d\\n", time);%}
                          COMPONENT c = Arm() AT (0,0,2) ABSOLUTE JUMP a WHEN (do_jump)
                          END
                          """)
        self._do_jump_tests(contents, 100)

    @compiled
    def test_split(self):
        contents = dedent("""\
                          define instrument test_split(int splits) trace
                          component a = Arm() AT (0,0,0) ABSOLUTE EXTEND %{printf("a\\n");%}
                          split splits component b = Arm() at (0,0,0) absolute extend %{printf("b\\n");%}
                          split splits component c = Arm() at (0,0,0) absolute extend %{printf("c\\n");%}
                          end 
                          """)
        instr = parse_mcstas_instr(contents)
        splits = 10
        results, data = compile_and_run(instr, f"-n 1 splits={splits}")
        lines = results.decode('utf-8').splitlines()
        self.assertEqual(1, sum(line == 'a' for line in lines))
        self.assertEqual(splits, sum(line == 'b' for line in lines))
        self.assertEqual(splits*splits, sum(line == 'c' for line in lines))
        # The *order* of output is known as well:
        expected = ['a', *(['b', *(['c'] * splits)] * splits)]
        for ex, ln in zip(expected, lines):
            self.assertEqual(ex, ln)

    @compiled
    def test_when(self):
        instr = parse_mcstas_instr(dedent("""\
            define instrument test_when(dummy) 
            declare %{int count;%}
            initialize %{count = 0;%}
            trace
            component a = Arm() at (0,0,0) absolute extend %{++count;%}
            component b = Arm() when (count%2==1) at (0,0,0) absolute extend %{printf("count=%d\\n", count);%}
            end 
        """))
        n_max = 10
        results, data = compile_and_run(instr, f"-n {n_max} dummy=1")
        lines = results.decode('utf-8').splitlines()
        print(lines)
        expected = [f'count={x}' for x in range(1, n_max, 2)]
        for ex, ln in zip(expected, lines):
            self.assertEqual(ex, ln)

    @compiled
    def test_group(self):
        instr = parse_mcstas_instr(dedent("""\
            define instrument test_group(dummy)
            declare %{int count;%}
            initialize %{count = -1;%}
            trace
            component o = Arm() at (0,0,0) absolute extend %{++count;%}
            component a = Arm() when (count%4==0) at (0,0,0) absolute group first 
                          extend %{printf("a=%d\\n", count); SCATTER; %}
            component b = Arm() when (count%2==0) at (0,0,0) absolute group first 
                          extend %{printf("b=%d\\n", count);%}
            end 
        """))
        n_max = 10
        results, data = compile_and_run(instr, f"-n {n_max} dummy=1")
        lines = results.decode('utf-8').splitlines()
        print(lines)
        expected = [f'{"a" if x % 4 == 0 else "b"}={x}' for x in range(0, n_max, 2)]
        for ex, ln in zip(expected, lines):
            self.assertEqual(ex, ln)


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations
from pathlib import Path
from mccode_antlr.reader import Registry

def regular_mccode_runtime_dict(args: dict) -> dict:
    def insert_best_of(src: dict, snk: dict, names: tuple):
        def get_best_of():
            for name in names:
                if name in src:
                    return src[name]
            raise RuntimeError(f"None of {names} found in {src}")

        if any(x in src for x in names):
            snk[names[0]] = get_best_of()
        return snk

    t = insert_best_of(args, {}, ('seed', 's'))
    t = insert_best_of(args, t, ('ncount', 'n'))
    t = insert_best_of(args, t, ('dir', 'out_dir', 'd'))
    t = insert_best_of(args, t, ('trace', 't'))
    t = insert_best_of(args, t, ('gravitation', 'g'))
    t = insert_best_of(args, t, ('bufsiz',))
    t = insert_best_of(args, t, ('format',))
    return t


def mccode_runtime_dict_to_args_list(args: dict) -> list[str]:
    """Convert a dictionary of McCode runtime arguments to a string.

    :parameter args: A dictionary of McCode runtime arguments.
    :return: A list of arguments suitable for use in a command line call to a McCode compiled instrument.
    """
    # convert to a standardized string:
    out = []
    if 'seed' in args and args['seed'] is not None:
        out.append(f'--seed={args["seed"]}')
    if 'ncount' in args and args['ncount'] is not None:
        out.append(f'--ncount={args["ncount"]}')
    if 'dir' in args and args['dir'] is not None:
        out.append(f'--dir={args["dir"]}')
    if 'trace' in args and args['trace']:
        out.append('--trace')
    if 'gravitation' in args and args['gravitation']:
        out.append('--gravitation')
    if 'bufsiz' in args and args['bufsiz'] is not None:
        out.append(f'--bufsiz={args["bufsiz"]}')
    if 'format' in args and args['format'] is not None:
        out.append(f'--format={args["format"]}')
    return out


def mccode_runtime_parameters(args: dict, params: dict) -> str:
    first = ' '.join(mccode_runtime_dict_to_args_list(args))
    second = ' '.join(f'{k}={v}' for k, v in params.items())
    return f'{first} {second}'


def sort_args(args: list[str]) -> list[str]:
    """Take the list of arguments and sort them into the correct order for McCode run."""
    # TODO this is a bit of a hack, but it works for now
    first, last = [], []
    k = 0
    while k < len(args):
        if args[k].startswith('-'):
            first.append(args[k])
            k += 1
            if '=' not in first[-1] and k < len(args) and not args[k].startswith('-') and '=' not in args[k]:
                first.append(args[k])
                k += 1
        else:
            last.append(args[k])
            k += 1
    return first + last


def mccode_run_script_parser(prog: str):
    from argparse import ArgumentParser
    from pathlib import Path

    def resolvable(name: str):
        return None if name is None else Path(name).resolve()

    parser = ArgumentParser(prog=prog, description=f'Convert and run mccode_antlr-3 instr and comp files to {prog} runtime in C')
    aa = parser.add_argument

    aa('filename', type=resolvable, nargs=1, help='.instr file name to be converted')
    aa('parameters', nargs='*', help='Parameters to be passed to the instrument', type=str, default=None)
    aa('-o', '--output-file', type=str, help='Output filename for C runtime binary', default=None)
    aa('-d', '--directory', type=str, help='Output directory for C runtime artifacts')
    aa('-I', '--search-dir', action='append', type=resolvable, help='Extra component search directory')
    aa('-t', '--trace', action='store_true', help="Enable 'trace' mode for instrument display")
    aa('-v', '--version', action='store_true', help='Print the McCode version')
    aa('--source', action='store_true', help='Embed the instrument source code in the executable')
    aa('--verbose', action='store_true', help='Verbose output')

    aa('-n', '--ncount', nargs=1, type=int, default=None, help='Number of neutrons to simulate')
    aa('-m', '--mesh', action='store_true', default=False, help='N-dimensional mesh scan')
    aa('-s', '--seed', nargs=1, type=int, default=None, help='Random number generator seed')
    aa('-g', '--gravitation', action='store_true', default=False,
       help='Enable gravitation for all trajectories')
    aa('--bufsiz', nargs=1, type=int, default=None, help='Monitor_nD list/buffer-size')
    aa('--format', nargs=1, type=str, default=None, help='Output data files using FORMAT')
    aa('--dryrun', action='store_true', default=False,
       help='Do not run any simulations, just print the commands')
    aa('--parallel', action='store_true', default=False, help='Use MPI multi-process parallelism')
    aa('--gpu', action='store_true', default=False, help='Use GPU OpenACC parallelism')
    aa('--process-count', nargs=1, type=int, default=0, help='MPI process count, 0 == System Default')

    return parser


def print_version_information():
    from mccode_antlr.version import version
    print(f'mccode_antlr code generator version {version()}')
    print(' Copyright (c) European Spallation Source ERIC, 2023-2024')
    print('Based on McStas/McXtrace version 3')
    print(' Copyright (c) DTU Physics and Risoe National Laboratory, 1997-2023')
    print(' Additions (c) Institut Laue Langevin, 2003-2019')
    print('All rights reserved\n\nComponents are (c) their authors, see component headers.')


def parse_mccode_run_script(prog: str):
    import sys
    from .range import parse_scan_parameters
    sys.argv[1:] = sort_args(sys.argv[1:])
    args = mccode_run_script_parser(prog).parse_args()
    parameters = parse_scan_parameters(args.parameters)
    return args, parameters


def mccode_compile(instr, directory, generator, target: dict | None = None, config: dict | None = None, **kwargs):
    from mccode_antlr.compiler.c import compile_instrument, CBinaryTarget
    from loguru import logger

    def_target = CBinaryTarget(mpi=False, acc=False, count=1, nexus=False)
    def_config = dict(default_main=True, enable_trace=False, portable=False, include_runtime=True,
                      embed_instrument_file=False, verbose=False)
    def_config.update(config or {})
    def_target.update(target or {})

    try:
        binary = compile_instrument(instr, def_target, directory, generator=generator, config=def_config, **kwargs)
    except RuntimeError as compilation_error:
        logger.error(f'Failed to compile instrument: {compilation_error}')
        raise compilation_error

    return binary, def_target


def mccode_run_compiled(binary, target, directory: Path | str, parameters: str, capture: bool = True, dry_run: bool = False):
    from mccode_antlr.compiler.c import run_compiled_instrument
    from mccode_antlr.loader import read_mccode_dat
    from pathlib import Path

    result = run_compiled_instrument(binary, target, f'--dir {directory} {parameters}', capture=capture, dry_run=dry_run)
    sim_files = list(Path(directory).glob('**/*.dat'))
    dats = {file.stem: read_mccode_dat(file) for file in sim_files}
    return result, dats


def mccode_run_scan(name: str, binary, target, parameters, directory, grid: bool, capture: bool = True, dry_run: bool = False, **r_args):
    from .range import parameters_to_scan
    n_pts, names, scan = parameters_to_scan(parameters, grid=grid)
    # n_zeros = len(str(n_pts))

    args = regular_mccode_runtime_dict(r_args)

    if directory is None:
        from datetime import datetime
        directory = Path(f'{name}{datetime.now().strftime("%Y%m%d_%H%M%S")}')

    # if there is only one point, we don't need to scan
    if n_pts > 1:
        directory.mkdir(parents=True, exist_ok=True)
        results = []
        for number, values in enumerate(scan):
            # TODO Use the following line instead of the one after it when McCode is fixed to use zero-padded folder names
            # # runtime_arguments['dir'] = args["dir"].joinpath(str(number).zfill(n_zeros))
            this_directory = directory.joinpath(str(number))
            pars = mccode_runtime_parameters(r_args, parameters)
            r, d = mccode_run_compiled(binary, target, this_directory, pars, capture=capture, dry_run=dry_run)
            results.append((r, d))
        return results
    else:
        directory.parent.mkdir(parents=True, exist_ok=True)
        print(parameters)
        pars = mccode_runtime_parameters(args, parameters)
        return mccode_run_compiled(binary, target, directory, pars, capture=capture, dry_run=dry_run)


def mccode_run(instrument, generator, parameters, directory: str | Path, binary_name: str | None = None,
               trace: bool = False, source: bool = False, verbose: bool = False,
               parallel: bool = False, gpu: bool = False, process_count: int = 0,
               mesh: bool = False, seed: int | None = None, ncount: int | None = None,
               gravitation: bool | None = None, bufsize: int | None = None, dryrun: bool = False, fmt: str | None = None,
               ):
    from os import access, R_OK
    from datetime import datetime
    if not isinstance(directory, Path):
        directory = Path(directory)
    if binary_name is not None:
        binary_path = directory.joinpath(binary_name)
    else:
        binary_path = directory.joinpath(instrument.name)

    if binary_path.exists() and not access(binary_path, R_OK):
        raise ValueError(f"{binary_path} exists but is not an executable")

    target = {'mpi': parallel, 'acc': gpu, 'count': process_count, 'nexus': False}
    if not binary_path.exists():
        config = {'enable_trace': trace, 'embed_instrument_file': source, 'verbose': verbose}
        binary_path, target = mccode_compile(instrument, binary_path, generator, target=target, config=config)

    runtime = dict(
        seed=seed,
        ncount=ncount,
        trace=trace,
        gravitation=gravitation,
        bufsiz=bufsize,
        format=fmt,
        dry_run=dryrun,
        capture=(not verbose) if verbose is not None else False,
    )
    out_dir = directory.joinpath(f'{instrument.name}{datetime.now().strftime("%Y%m%d_%H%M%S")}')

    mccode_run_scan(instrument.name, binary_path, target, parameters, out_dir, mesh, **runtime)


def mccode_run_cmd(flavor: str, registry: Registry, generator: dict):
    from pathlib import Path
    from mccode_antlr.reader import Reader
    from mccode_antlr.reader import LocalRegistry
    from os import R_OK, access

    args, parameters = parse_mccode_run_script(flavor)
    filename = args.filename if isinstance(args.filename, str) else next(iter(args.filename))
    config = dict(
        enable_trace=args.trace if args.trace is not None else False,
        embed_instrument_file=args.source if args.source is not None else False,
        verbose=args.verbose if args.verbose is not None else False,
        output=args.output_file if args.output_file is not None else filename.with_suffix('.c')
    )
    target = dict(
        mpi=args.parallel,
        acc=args.gpu,
        count=args.process_count,
        nexus=False
    )
    runtime = dict(
        seed=args.seed[0] if args.seed is not None else None,
        ncount=args.ncount[0] if args.ncount is not None else None,
        trace=args.trace,
        gravitation=args.gravitation,
        bufsiz=args.bufsiz[0] if args.bufsiz is not None else None,
        format=args.format[0] if args.format is not None else None,
        dry_run=args.dryrun,
        capture=(not args.verbose) if args.verbose is not None else False,
    )
    # check if the filename is actually a compiled instrument already:
    if args.output_file is None and filename.exists() and access(filename, R_OK):
        binary = filename
        name = filename.stem
    else:
        # McCode always requires access to a remote Pooch repository:
        registries = [registry]
        # A user can specify extra (local) directories to search for included files using -I or --search-dir
        if args.search_dir is not None and len(args.search_dir):
            registries.extend([LocalRegistry(d.stem, d) for d in args.search_dir])
        # And McCode-3 users expect to always have access to files in the current working directory
        registries.append(LocalRegistry('working_directory', f'{Path().resolve()}'))
        # Construct the object which will read the instrument and component files, producing Python objects
        reader = Reader(registries=registries)
        # Read the provided .instr file, including all specified .instr and .comp files along the way
        instrument = reader.get_instrument(filename)
        name = instrument.name
        # Generate the C binary for the instrument -- will output to, e.g., {instrument.name}.out, in the current directory
        # unless if output_file was specified
        binary, target = mccode_compile(instrument, args.output_file, generator, target=target, config=config)

    if not len(parameters):
        from loguru import logger
        logger.error("Interactive parameter entry does not currently work")
        logger.info(f"Execute `{binary} --list-parameters` to check expected parameters")
        return

    mccode_run_scan(name, binary, target, parameters, args.directory, args.mesh, **runtime)


def mcstas_cmd():
    from mccode_antlr.reader import MCSTAS_REGISTRY
    from mccode_antlr.translators.target import MCSTAS_GENERATOR
    mccode_run_cmd('mcstas', MCSTAS_REGISTRY, MCSTAS_GENERATOR)


def mcxtrace_cmd():
    from mccode_antlr.reader import MCXTRACE_REGISTRY
    from mccode_antlr.translators.target import MCXTRACE_GENERATOR
    mccode_run_cmd('mcxtrace', MCXTRACE_REGISTRY, MCXTRACE_GENERATOR)


def mcstas_run(instrument, directory):
    from mccode_antlr.reader import MCSTAS_REGISTRY
    from mccode_antlr.translators.target import MCSTAS_GENERATOR
    mccode_run(instrument, directory, MCSTAS_REGISTRY, MCSTAS_GENERATOR)


def mcxtrace_run(instrument, directory):
    from mccode_antlr.reader import MCXTRACE_REGISTRY
    from mccode_antlr.translators.target import MCXTRACE_GENERATOR
    mccode_run(instrument,  directory, MCXTRACE_REGISTRY, MCXTRACE_GENERATOR)



def cogen_raytrace(source, ok_to_skip):
    lines = [
        "/* *****************************************************************************",
        f"* instrument '{source.name}' TRACE",
        "***************************************************************************** */",
        "",
        "#ifndef FUNNEL",
        "#pragma acc routine",
        "int raytrace(_class_particle* _particle) { "
        f"/* single event propagation, called by mccode_main for {source.name}:TRACE */",
        "",
        "  /* init variables and counters for TRACE */",
        #  // we need this override, since "comp" is not defined in raytrace() - see section-wide define
        "  #undef ABSORB0",
        "  #undef ABSORB",
        "  #define ABSORB0 do { DEBUG_ABSORB(); MAGNET_OFF; ABSORBED++;} while(0)",

        "  #define ABSORB ABSORB0",
        #  /* Debugging (initial state). */
        "  DEBUG_ENTER();",
        "  DEBUG_STATE();",
    ]
    print("\n-----------------------------------------------------------")
    print("\nGenerating single GPU kernel or single CPU section layout:")

    for name, target in [(c.name, j.target) for c in source.components for j in c.jump if j.iterate]:
        lines.append(f'  _particle->_logic.Jump_{name}_{target}=0;')

    lines.extend([
        "  _particle->flag_nocoordschange=0; /* Init */",
        "  _class_particle _particle_save;",
        "  /* the main iteration loop for one incoming event */",
        "  while (!ABSORBED) { /* iterate event until absorbed */",
        "    /* send particle event to component instance, one after the other */",
    ])

    # /* now we produce the list of statements for each component index, with the attached logic */
    for index, comp in enumerate(source.components):
        if comp.split is not None:
            # split is a Value, containing a number or expression indicating *how many* split particles to produce
            print(f'-> SPLIT {comp.split:p} at component {comp.name}')
            lines.extend([
                '#ifndef NOSPLIT',
                f'    /* start SPLIT at {comp.name} */',
                '    if (!ABSORBED) {',
                f'    _class_particle Split_{comp.name}_particle = *_particle;',
                f'    int SplitS_{comp.name} = {comp.split:p};',
                # following https://github.com/McStasMcXtrace/McCode/commit/72cb740d465e2780b228bd937cf62429a96f64d5
                '    #pragma acc loop independent',
                f'    for (int SplitN_{comp.name}=0; SplitN_{comp.name} < SplitS_{comp.name}; SplitN_{comp.name}++) {{',
                '    randstate_t randbackup = *_particle->randstate;',
                f'    *_particle=Split_{comp.name}_particle;',
                '    *_particle->randstate = randbackup;',
                f'    p /= SplitS_{comp.name} > 0 ? SplitS_{comp.name} : 1;',
                '#endif'
            ])
        # Coordinate transforms WRT PREVIOUS
        lines.append(f'    /* begin component {comp.name}={comp.type.name}() [{1+index}] */')
        if not ok_to_skip[index]:
            lines.extend([
                "    if (!_particle->flag_nocoordschange) { // flag activated by JUMP to pass coords change",
                f"      if (_{comp.name}_var._rotation_is_identity) {{",
                f"        if(!_{comp.name}_var._position_relative_is_zero) {{",
                f"          coords_get(coords_add(coords_set(x,y,z), _{comp.name}_var._position_relative),&x, &y, &z);",
                "        }",
                "      } else {",
                f"          mccoordschange(_{comp.name}_var._position_relative, _{comp.name}_var._rotation_relative, _particle);",
                "      }",
                "    }",
            ])
        lines.extend([
            f'    if (!ABSORBED && _particle->_index == {1+index}) {{',
            '      _particle->flag_nocoordschange=0; /* Reset if we came here from a JUMP */',
        ])
        if comp.group is None or source.groups[comp.group].is_leader(index):
            # This allows an empty ('') named group :/
            if not ok_to_skip[index]:
                lines.append('      _particle_save = *_particle;')
        else:
            lines.extend([
                '      // 2nd or higher GROUP member, "reuse" coordinate-changed _particle_Save from 1st GROUP element.',
                f'      mccoordschange(_{comp.name}_var._position_relative, _{comp.name}_var._rotation_relative, &_particle_save);',
            ])
        if not ok_to_skip[index]:
            lines.extend([
                f'      DEBUG_COMP(_{comp.name}_var._name);',
                '      DEBUG_STATE();'
            ])
        if len(comp.type.trace) or len(comp.extend):
            # WHEN
            if comp.when is not None:
                lines.append(f'      if (({comp.when:p}))  // conditional WHEN execution')
            lines.extend([
                f'      class_{comp.type.name}_trace(&_{comp.name}_var, _particle);{" /* contains EXTEND code */" if len(comp.extend) else ""}',
                '      if (_particle->_restore)',
                '        particle_restore(_particle, &_particle_save);'
            ])
        # If we have a JUMP, change the index
        for jump in comp.jump:
            if jump.iterate:
                lines.extend([
                    f'      if (++_particle->_logic.Jump_{comp.name}_{jump.target} < {jump.condition:p}) {{ /* test for iteration */',
                    f'        _particle->_index = {jump.absolute_target};',  # NO -1 despite to ++ below, 0-based target
                    '        _particle->flag_nocoordschange=1; /* pass coordinate transformations when jumping */',
                    '      } else {',
                    f'        _particle->_logic.Jump_{comp.name}_{jump.target} = 0; /* reset Jump top and go forward */',
                    '      }'
                ])
            else:
                lines.extend([
                    f'      if ({jump.condition:p}) {{/* conditional JUMP to {jump.target} */',
                    f'        _particle->_index = {jump.absolute_target};',  # target is 0-based, _index 1; OK
                    '        _particle->flag_nocoordschange=1; /* pass coordinate transformations when jumping */',
                    '      }'
                ])
        # If we are in a group handle SCATTERED
        if comp.group is not None:
            group = source.groups[comp.group]
            fn = group.first.name
            ln = group.last.name
            lid = group.last_id
            lines.extend([
                f'      // GROUP {group.name}: from {fn} [{group.first_id}] to {ln} [{1+lid}]',
                # Skip over when SCATTERED in the group:
                f'      if (SCATTERED) _particle->_index = {1+lid}; '
                '// when SCATTERED in GROUP: reach exit of GROUP after {ln}',
                #
            ])
            if index == lid:
                lines.append('      else ABSORBED=1; // Not SCATTERED by end of GROUP: particle does not progress')
            else:
                lines.append('      else particle_restore(_particle, &_particle_save); // not SCATTERED in GROUP, restore')

        lines.append('      _particle->_index++;')
        if not ok_to_skip[index]:
            lines.append('      if (!ABSORBED) { DEBUG_STATE(); }')
        lines.append(f'    }} /* end of component {comp.name} [{1+index}] */')

    # /* now we close the SPLIT loops, unrolled from last to 1st */
    for comp in reversed(source.components):
        if comp.split is not None:
            lines.extend([
                '#ifndef NOSPLIT',
                f'    }} /* end SPLIT at {comp.name} */',
                f'    }} /* if (!ABSORBED) relating to SPLIT at {comp.name} */',
                '#endif'
            ])

    # /* handle ABSORB when no more comp, go to next comp, and RESTORE */
    lines.extend([
        # /* propagate to the next component */
        f'    if (_particle->_index > {len(source.components)})',
        # /* if we reach the last component and nothing happened, ABSORB */
        '      ABSORBED++; /* absorbed when passed all components */',
        "  } /* while !ABSORBED */",
        "",
        "  DEBUG_LEAVE()",
        "  particle_restore(_particle, &_particle_save);",
        #  /* Debugging (final state). */
        "  DEBUG_STATE()",
        "",
        "  return(_particle->_index);",
        "} /* raytrace */",
        #  // write the "raytrace_all()" code (for loop previously in mccode_main.c).
        "",
        "/* loop to generate events and call raytrace() propagate them */",
        "void raytrace_all(unsigned long long ncount, unsigned long seed) {",
        "",
        "  /* CPU-loop */",
        "  unsigned long long loops;",
        "  loops = ceil((double)ncount/gpu_innerloop);",
        "  /* if on GPU, printf has been globally nullified, re-enable here */",
        "  #ifdef OPENACC",
        "  #undef strlen",
        "  #undef strcmp",
        "  #undef exit",
        "  #undef printf",
        "  #undef sprintf",
        "  #undef fprintf",
        "  #endif",
        "",
        "  #ifdef OPENACC",
        "  if (ncount>gpu_innerloop) {",
        '    printf("Defining %llu CPU loops around GPU kernel and adjusting ncount\\n",loops);',
        "    mcset_ncount(loops*gpu_innerloop);",
        "  } else {",
        "    #endif",
        "    loops=1;",
        "    gpu_innerloop = ncount;",
        "    #ifdef OPENACC",
        "  }",
        "    #endif",
        "",
        "  for (unsigned long long cloop=0; cloop<loops; cloop++) {",
        "    #ifdef OPENACC",
        '    if (loops>1) fprintf(stdout, "%d..", (int)cloop); fflush(stdout);',
        "    #endif",
        "",
        "    /* if on GPU, re-nullify printf */",
        "    #ifdef OPENACC",
        "    #undef strlen",
        "    #undef strcmp",
        "    #undef exit",
        "    #undef printf",
        "    #undef sprintf",
        "    #undef fprintf",
        "    #endif",
        "",
        "    #pragma acc parallel loop num_gangs(numgangs) vector_length(vecsize)",
        "    for (unsigned long pidx=0 ; pidx < gpu_innerloop ; pidx++) {",
        "      _class_particle particleN = mcgenstate(); // initial particle",
        "      _class_particle* _particle = &particleN;",
        "      particleN._uid = pidx;",
        "      #ifdef USE_MPI",
        "      particleN._uid += mpi_node_rank * ncount;",
        "      #endif",
        "",
        "      srandom(_hash((pidx+1)*(seed+1)));",
        "      particle_uservar_init(_particle);",
        "",
        "      raytrace(_particle);",
        "    } /* inner for */",
        "    seed = seed+gpu_innerloop;",
        "  } /* CPU for */",
        "  /* if on GPU, printf has been globally nullified, re-enable here */",
        "  #ifdef OPENACC",
        "  #undef strlen",
        "  #undef strcmp",
        "  #undef exit",
        "  #undef printf",
        "  #undef sprintf",
        "  #undef fprintf",
        "  #endif",
        "  MPI_MASTER(",
        '  printf("*** TRACE end *** \\n");',
        "  );",
        "} /* raytrace_all */",
        "",
        "#endif //no-FUNNEL",
    ])

    return '\n'.join(lines)


def cogen_funnel(source, ok_to_skip):
    from ..common.utilities import escape_str_for_c
    print('\n-----------------------------------------------------------')
    print('\nGenerating GPU/CPU -DFUNNEL layout:')

    lines = [
        "",
        "#ifdef FUNNEL",
        "// Alternative raytrace algorithm which iterates all particles through",
        "// one component at the time, can remove absorbs from the next loop and",
        "// switch between cpu/gpu.",
        "void raytrace_all_funnel(unsigned long long ncount, unsigned long seed) {",
        "",
        "  // set up outer (CPU) loop / particle batches",
        "  unsigned long long loops;",
        "",
        "  /* if on GPU, printf has been globally nullified, re-enable here */",
        "  #ifdef OPENACC",
        "  #undef strlen",
        "  #undef strcmp",
        "  #undef exit",
        "  #undef printf",
        "  #undef sprintf",
        "  #undef fprintf",
        "  #endif",
        "",
    ]

    for i, c in [(i, c) for i, c in enumerate(source.components) if len(c.jump)]:
        message = [
            f'\nWARNING:\n --> JUMP found at COMPONENT {i}, {c.name}',
            ' --> JUMPS are not supported in FUNNEL mode and are ignored',
            ' --> Your instrument may give different output with FUNNEL'
        ]
        print('\n'.join(message))
        lines.extend(f'printf("{escape_str_for_c(ml)}\\n");' for ml in message)

    lines.extend([
        "  #ifdef OPENACC",
        "  loops = ceil((double)ncount/gpu_innerloop);",
        "  if (ncount>gpu_innerloop) {",
        '    printf("Defining %llu CPU loops around kernel and adjusting ncount\\n",loops);',
        "    mcset_ncount(loops*gpu_innerloop);",
        "  } else {",
        "  #endif",
        "    loops=1;",
        "    gpu_innerloop = ncount;",
        "  #ifdef OPENACC",
        "  }",
        "  #endif",
        "",
        "  // create particles struct and pointer arrays (same memory used by all batches)",
        "  _class_particle* particles = malloc(gpu_innerloop*sizeof(_class_particle));",
        "  _class_particle* pbuffer = malloc(gpu_innerloop*sizeof(_class_particle));",
        "  long livebatchsize = gpu_innerloop;",
        "",
        #  // we need this override, since "comp" is not defined in raytrace() - see section-wide define
        "  #undef ABSORB0",
        "  #undef ABSORB",
        "  #define ABSORB0 do { DEBUG_ABSORB(); MAGNET_OFF; ABSORBED++; } while(0)",
        "  #define ABSORB ABSORB0",
        #  // batches
        "  // outer loop / particle batches",
        "  for (unsigned long long cloop=0; cloop<loops; cloop++) {",
        '    if (loops>1) fprintf(stdout, "%d..", (int)cloop); fflush(stdout);',
        "",
        #  // init batch
        "    // init particles",
        "    #pragma acc parallel loop present(particles[0:livebatchsize])",
        "    for (unsigned long pidx=0 ; pidx < livebatchsize ; pidx++) {",
        "      // generate particle state, set loop index and seed",
        "      particles[pidx] = mcgenstate();",
        "      _class_particle* _particle = particles + pidx;",
        "      _particle->_uid = pidx;",
        "      #ifdef USE_MPI",
        "      _particle->_uid += mpi_node_rank * ncount;",
        "      #endif",
        "      srandom(_hash((pidx+1)*(seed+1))); // _particle->state usage built into srandom macro",
        "      particle_uservar_init(_particle);",
        "    }",
        "",
        '    // iterate components'
    ])
    cpu_last = not source.components[0].cpu if len(source.components) else False
    for index, comp in enumerate(source.components):
        if not comp.type.acc:
            print(f'Component {comp.name} is NOACC, CPUONLY={comp.cpu}')
            print('->FUNNEL mode enabled, SPLIT within buffer.')
            lines.append('        #define JUMP_FUNNEL')
        if index > 0 and (comp.cpu != cpu_last or comp.split is not None):
            lines.append("    }")
            if comp.split is not None:
                lines.extend([
                    '',
                    "    // SPLIT with available livebatchsize ",
                    f"    long mult_{comp.name};",
                    f"    livebatchsize = sort_absorb_last(particles, pbuffer, livebatchsize, gpu_innerloop, 1, &mult_{comp.name});",
                    "    //printf(\"livebatchsize: %ld, split: %ld\\n\",  livebatchsize, mult);",
                ])
                print(f'-> SPLIT within buffer at component {comp.name}')

        elif comp.cpu:
            print(f'-> CPU section from component {comp.name}')
        else:
            print(f'-> GPU kernel from component {comp.name}')

        if index == 0 or comp.cpu != cpu_last or comp.split is not None:
            lines.append("")
            if not comp.cpu:
                lines.extend([
                    "    #pragma acc parallel loop present(particles[0:livebatchsize])"
                ])
            else:
                lines.extend([
                    "    #ifdef MULTICORE",
                    "    #pragma acc parallel loop device_type(host)"
                    "    #endif"
                ])
            lines.extend([
                "    for (unsigned long pidx=0 ; pidx < livebatchsize ; pidx++) {",
                "      _class_particle* _particle = &particles[pidx];",
                "      _class_particle _particle_save;",
            ])

        lines.extend([
            f'      // {comp.name}',
            f'      if (!ABSORBED && _particle->_index == {index+1}) {{',
        ])
        if not ok_to_skip[index]:
            lines.extend([
                "        #ifndef MULTICORE",
                f"        if (_{comp.name}_var._rotation_is_identity) {{",
                f"          coords_get(coords_add(coords_set(x,y,z), _{comp.name}_var._position_relative),&x, &y, &z);",
                "        } else",
                "        #endif",
                "        {",
                f"          mccoordschange(_{comp.name}_var._position_relative, _{comp.name}_var._rotation_relative, _particle);",
                '        }',
                "        _particle_save = *_particle;",
            ])
        if len(comp.type.trace) or len(comp.extend):
            if comp.when is not None:
                lines.append(f'    if (({comp.when:p}))  // conditional WHEN')
            lines.extend([
                f'        class_{comp.type.name}_trace(&_{comp.name}_var, _particle);{" /* contains EXTEND code */" if len(comp.extend) else ""}',
                '         if (_particle->_restore) particle_restore(_particle, &_particle_save);'
            ])
        if comp.group is not None:
            group = source.groups[comp.group]
            if len(group.ids) < 2:
                print(f'\n!!! WARNING: GROUP {group.name} seems to include only one COMPONENT:')
                print(f'!!!   --> {comp.name} <--')
                print('!!! This may lead to unphysical simulation behaviour!')
            ln = group.last.name
            lid = group.last_id
            lines.extend([
                f'    // GROUP {group.name}: from {group.first.name} [{group.first_id}] to {ln} [{lid}]',
                # Skip over when SCATTERED in the group:
                f'    if (SCATTERED) _particle->_index = {lid}; '
                '// when SCATTERED in GROUP: reach exit of GROUP after {ln}',
                #
            ])
            if index == lid:
                lines.append('    else ABSORBED=1; // Not SCATTERED by end of GROUP: removes left events')
            else:
                lines.append("      else ABSORBED=0; // not SCATTERED within GROUP: always tries next")

        lines.append("        _particle->_index++;")
        lines.append("      }")
        cpu_last = comp.cpu

    lines.extend([
        '    }',
        "    // jump to next viable seed",
        "    seed = seed + gpu_innerloop;",
        "  } // outer loop / particle batches",
        "",
        "  free(particles);",
        "  free(pbuffer);",
        "",
        '  printf("\\n");',
        "} /* raytrace_all_funnel */",
        "#endif // FUNNEL",
        "",
    ])

    print("\n-----------------------------------------------------------")

    return '\n'.join(lines)



"""Data structures required for representing the contents of a McCode comp file"""
from dataclasses import dataclass, field
from ..common import ComponentParameter, MetaData, parameter_name_present, RawC, blocks_to_raw_c


# Could this be replaced by a subclassed 'name' class? E.g., Slit.comp <=> class McCompSlit(McComp)?
@dataclass
class Comp:
    """Intermediate representation of a McCode component definition

    Read from a .comp file
    For output to a runtime source file
    """
    name: str = None           # Component *type* name, e.g. {name}.comp
    category: str = None       # Component type catagory -- nearly free-form
    define: tuple[ComponentParameter, ...] = field(default_factory=tuple)   # C #define'ed parameters
    setting: tuple[ComponentParameter, ...] = field(default_factory=tuple)  # Formal 'setting' parameters
    output: tuple[ComponentParameter, ...] = field(default_factory=tuple)   # 'output' parameters
    metadata: tuple[MetaData, ...] = field(default_factory=tuple)           # metadata for use by simulation consumers
    dependency: str = None     # compile-time dependency
    acc: bool = True           # False if this component *can not* work under OpenACC
    # literal strings writen into C source files
    share: tuple[RawC, ...] = field(default_factory=tuple)       # function(s) for all instances of this class
    user: tuple[RawC, ...] = field(default_factory=tuple)        # struct members for _particle
    declare: tuple[RawC, ...] = field(default_factory=tuple)     # global parameters used in component trace
    initialize: tuple[RawC, ...] = field(default_factory=tuple)  # initialization of global declare parameters
    trace: tuple[RawC, ...] = field(default_factory=tuple)       # per-particle interaction executed at TRACE time
    save: tuple[RawC, ...] = field(default_factory=tuple)        # statements executed after TRACE to save results
    final: tuple[RawC, ...] = field(default_factory=tuple)       # clean-up memory for global declare parameters
    display: tuple[RawC, ...] = field(default_factory=tuple)     # draw this component

    def __hash__(self):
        return hash(repr(self))

    # @property
    # def parameters(self):
    #     return self.define + self.setting

    def has_parameter(self, name: str):
        return parameter_name_present(self.define, name) or parameter_name_present(self.setting, name)

    def get_parameter(self, name: str, default=None):
        if parameter_name_present(self.define, name):
            return [par for par in self.define if par.name == name][0]
        if parameter_name_present(self.setting, name):
            return [par for par in self.setting if par.name == name][0]
        return default

    def compatible_parameter_value(self, name: str, value):
        return self.get_parameter(name).compatible_value(value)

    def parameter_name_used(self, parameter_type: str, name: str):
        """Verify that a new parameter name is not already in use."""
        if parameter_name_present(self.define, name):
            raise RuntimeError(f"{parameter_type} parameter {name} is already an instance parameter!")
        if parameter_name_present(self.setting, name):
            raise RuntimeError(f"{parameter_type} parameter {name} is already a setting parameter!")
        if parameter_name_present(self.output, name):
            raise RuntimeError(f"{parameter_type} parameter {name} is already an output parameter!")

    def add_define(self, a: ComponentParameter):
        self.parameter_name_used('DEFINE', a.name)
        self.define += (a, )

    def add_setting(self, a: ComponentParameter):
        self.parameter_name_used('SETTING', a.name)
        self.setting += (a, )

    def add_output(self, a: ComponentParameter):
        # TODO in McCode-3 OUTPUT PARAMETERS are parsed but unused in the code-generator.
        #      And they can apparently anyway specify the same name as SETTING (and possibly DEFINE) PARAMETERS
        #      So only ensure that the same name isn't specified twice in OUTPUT, and consider dropping entirely.
        #self.parameter_name_used('OUTPUT', a.name)
        if parameter_name_present(self.output, a.name):
            raise RuntimeError(f"OUTPUT parameter {a.name} is already an output parameter!")
        self.output += (a, )

    def no_acc(self):
        self.acc = False

    # Meta programming these block setting methods is more pain than its worth:
    def SHARE(self, *blocks):
        self.share += blocks_to_raw_c(*blocks)

    def USERVARS(self, *blocks):
        self.user += blocks_to_raw_c(*blocks)

    def DECLARE(self, *blocks):
        self.declare += blocks_to_raw_c(*blocks)

    def INITIALIZE(self, *blocks):
        self.initialize += blocks_to_raw_c(*blocks)

    def TRACE(self, *blocks):
        self.trace += blocks_to_raw_c(*blocks)

    def SAVE(self, *blocks):
        self.save += blocks_to_raw_c(*blocks)

    def FINALLY(self, *blocks):
        self.final += blocks_to_raw_c(*blocks)

    def DISPLAY(self, *blocks):
        self.display += blocks_to_raw_c(*blocks)

    def add_metadata(self, m: MetaData):
        if any([x.name == m.name for x in self.metadata]):
            self.metadata = tuple([x for x in self.metadata if x.name != m.name])
        self.metadata += (m, )

    def collect_metadata(self):
        return self.metadata



from __future__ import annotations
from pathlib import Path
from loguru import logger
from dataclasses import dataclass, field

from .registry import Registry, MCSTAS_REGISTRY, registries_match, registry_from_specification
from ..comp import Comp
from ..common import Mode

def make_reader_error_listener(super_class, filetype, name, source, pre=5, post=2):
    class ReaderErrorListener(super_class):
        def __init__(self):
            self.filetype = filetype
            self.name = name
            self.source = source
            self.pre = pre
            self.post = post

        def syntaxError(self, recognizer, offendingSymbol, *args, **kwargs):
            if len(args) == 4 and isinstance(args[3], str):
                # The 'speedy-antlr' syntax
                char_index, line, column, msg = args
            else:
                # the antlr4 (4.13.0) syntax
                line, column, msg, e = args
            logger.error(f'Syntax error in parsing {self.filetype} {self.name} at {line},{column}')
            lines = self.source.split('\n')
            pre_lines = lines[line-self.pre:line]
            post_lines = lines[line:line+self.post]
            for line in pre_lines:
                logger.info(line)
            logger.error('~'*column + '^ ' + msg)
            for line in post_lines:
                logger.info(line)

    return ReaderErrorListener()


@dataclass
class Reader:
    registries: list[Registry] = field(default_factory=list)
    components: dict[str, Comp] = field(default_factory=dict)
    c_flags: list[str] = field(default_factory=list)

    def __post_init__(self):
        from .registry import ordered_registries
        if len(self.registries) == 0:
            self.registries = [MCSTAS_REGISTRY, ]
        self.registries = list(ordered_registries(self.registries))

    def prepend_registry(self, reg: Registry):
        self.registries[:0] = [reg, ]

    def append_registry(self, reg: Registry):
        self.registries.append(reg)

    def handle_search_keyword(self, spec: str):
        if not any(registries_match(reg, spec) for reg in self.registries):
            reg = registry_from_specification(spec)
            if reg is not None:
                self.prepend_registry(reg)
            else:
                raise RuntimeError(f"Registry specification {spec} did not specify a valid registry!")

    def add_c_flags(self, flags):
        self.c_flags.append(flags)

    def locate(self, name: str, which: str = None, ext: str = None, strict: bool = False):
        registries = self.registries if which is None else [x for x in self.registries if x.name in which]
        for reg in registries:
            if reg.known(name, ext, strict=strict):
                return reg.path(name, ext)
        names = [reg.name for reg in registries]
        msg = "registry " + names[0] if len(names) == 1 else 'registries: ' + ','.join(names)
        raise RuntimeError(f'{name} not found in {msg}')

    def contents(self, name: str, which: str = None, ext: str = None, strict: bool = False):
        registries = self.registries if which is None else [x for x in self.registries
                                                            if x.name in which]
        for reg in registries:
            if reg.known(name, ext, strict=strict):
                return reg.contents(name, ext)
        names = [reg.name for reg in registries]
        msg = "registry " + names[0] if len(names) == 1 else 'registries: ' + ','.join(
            names)
        raise RuntimeError(f'{name} not found in {msg}')

    def fullname(self, name: str, which: str = None, ext: str=None, strict: bool = False):
        registries = self.registries if which is None else [x for x in self.registries if x.name in which]
        for reg in registries:
            if reg.known(name, ext, strict=strict):
                return reg.fullname(name, ext)
        names = [reg.name for reg in registries]
        msg = "registry " + names[0] if len(names) == 1 else 'registries: ' + ','.join(names)
        raise RuntimeError(f'{name} not found in {msg}')

    def known(self, name: str, which: str = None, strict: bool = False):
        registries = self.registries if which is None else [x for x in self.registries if x.name in which]
        return any([reg.known(name, strict=strict) for reg in registries])

    def unique(self, name: str, which: str = None):
        registries = self.registries if which is None else [x for x in self.registries if x.name in which]
        return sum([1 for reg in registries if reg.unique(name)]) == 1

    def contain(self, name: str, which: str = None, strict: bool = False):
        registries = self.registries if which is None else [x for x in self.registries if x.name in which]
        return [reg.name for reg in registries if reg.known(name, strict=strict)]

    def stream(self, name: str, which: str = None, strict: bool = False):
        from antlr4 import InputStream
        return InputStream(self.contents(name, which=which, strict=strict))

    def add_component(self, name: str, current_instance_name=None):
        if name in self.components:
            raise RuntimeError("The named component is already known.")
        from antlr4 import InputStream
        from ..grammar import McComp_ErrorListener, McComp_parse
        from ..comp import CompVisitor
        source = self.contents(name, ext='.comp', strict=True)
        filename = str(self.locate(name, ext='.comp', strict=True))

        stream = InputStream(source)
        error_listener = make_reader_error_listener(McComp_ErrorListener, 'Component', name, source)
        tree = McComp_parse(stream, 'prog', error_listener)

        visitor = CompVisitor(self, filename, instance_name=current_instance_name)  # The visitor needs to be able to call *this* method
        res = visitor.visitProg(tree)
        if not isinstance(res, Comp):
            raise RuntimeError(f'Parsing component file {filename} did not produce a component object!')
        if res.category is None:
            # guess the component category from the registered filename (not fully resolved path)
            fullname = self.fullname(name, ext='.comp', strict=True)
            fullname = fullname if isinstance(fullname, Path) else Path(fullname)
            # if fullname is an absolute path, it comes from a local repository -- so we don't know what to do
            res.category = 'UNKNOWN' if fullname.is_absolute() else fullname.parts[0]
        self.components[name] = res

    def get_component(self, name: str, current_instance_name=None):
        if name not in self.components:
            self.add_component(name, current_instance_name=current_instance_name)
        return self.components[name]

    def get_instrument(self, name: str | Path, destination=None, mode: Mode | None = None):
        """Load and parse an instr Instrument definition file

        In McCode3 fashion, the instrument file *should* be in the current working directory.
        In new-fashion, the registry/registries will be checked if it is not.
        """
        from antlr4 import InputStream
        from ..grammar import McInstr_parse, McInstr_ErrorListener
        from ..instr import InstrVisitor, Instr
        path = name if isinstance(name, Path) else Path(name)
        if path.suffix != '.instr':
            path = path.with_suffix(f'{path.suffix}.instr')
        if path.exists() and path.is_file():
            source = path.read_text()
        else:
            path = self.locate(path.name)  # include the .instr for the search
            source = self.contents(path.name)
        filename = str(path.resolve())

        stream = InputStream(source)
        error_listener = make_reader_error_listener(
            McInstr_ErrorListener, 'Instrument', name, source
        )
        tree = McInstr_parse(stream, 'prog', error_listener)

        visitor = InstrVisitor(self, filename, destination=destination, mode=mode)
        res = visitor.visitProg(tree)
        if not isinstance(res, Instr):
            raise RuntimeError(f'Parsing instrument file {filename} did not produce an Instr object')
        res.source = filename
        res.flags = tuple(self.c_flags)
        res.registries = tuple(self.registries)
        return res



"""Translates a McComp instrument from its intermediate form to a C runtime source file."""
from loguru import logger
from dataclasses import dataclass
from ..reader import LIBC_REGISTRY
from .target import TargetVisitor
from .c_listener import extract_c_declared_variables


@dataclass
class CInclude:
    parent: str
    name: str
    order: int
    content: str = ''
    root: str = ''

    def __hash__(self):
        return hash(self.parent + self.name + self.content + self.root)

    def __str__(self):
        return f'{self.root}:{self.parent}({self.order}):{self.name}'

    def __lt__(self, other):
        if self.parent == other.parent:
            return self.order > other.order
        if self.root or other.root:
            if self.parent == self.root or self.parent == other.root:
                return True
            if other.parent == self.root or other.parent == other.root:
                return False
        if self.parent == other.name:
            return False
        if other.parent == self.name:
            return True
        return self.name < other.name


def sort_include_hierarchy(includes: list[CInclude]):
    if len(includes) == 0:
        return []
    logger.debug(f'sort includes {" ".join(str(x) for x in includes)}')
    # find the 'root' of the tree, a parent which is _not_ a named include:
    roots = list(set([x.parent for x in includes]).difference(set([x.name for x in includes])))
    options = {}
    for root in roots:
        for include in includes:
            include.root = root
        # Sort the includes such that if A includes B and C, and C includes B, then we get [C:B, A:C, A:B]
        # Or a more complicated case: G:(Z,A,W), Z:(B,Y), A:X, W(Y) -> [Z:Y, W:Y, A:X, B:W, Z:B, G:A, G:Z, G:W]
        used, output = [], []
        for include in sorted(includes, reverse=True):
            if include.name not in used:
                output.append(include)
                used.append(include.name)
        # Reduce output to [C:B, A:C], or [Z:Y, A:X, B:W, Z:B, G:A, G:Z], which would include the libraries correctly
        # Double check, that the include order does not _directly_ contradict one of the include directives:
        order = [x.name for x in output]
        contradicts = False
        for include in includes:
            if include.parent in order and include.name in order:
                parent_at = [index for index, name in enumerate(order) if name == include.parent][0]
                name_at = [index for index, name in enumerate(order) if name == include.name][0]
                if name_at > parent_at:
                    contradicts = True
        if not contradicts:
            logger.debug(f"sorted to {'  '.join(order)}")
            options[root] = output
    if len(options) == 0:
        raise RuntimeError(f'No valid sorting of {includes=}')
    # Maybe there's a case where the specified root matters?
    return list(options.values())[0]


class CTargetVisitor(TargetVisitor, target_language='c'):
    def _file_contents(self, filename, preserve_escapes=False):
        # First try the McCode C runtime libraries:
        path = self.include_path(filename)
        # Then any registries used in reading the file(s)
        if not path.is_file() and self.known(filename):
            path = self.locate(filename)
        if not path.is_file():
            raise RuntimeError(f"Can not include the file {filename} because it is not locatable")
        with path.open('r') as file:
            lines = [line.strip('\n') for line in file.readlines()]
        if preserve_escapes:
            lines = [line.encode('unicode-escape').decode('utf-8') for line in lines]
        return '\n'.join(lines)

    def _handle_raw_c_include(self, parent: str, raw_c: str):
        import re
        re_lib = re.compile(r'^\s*%include\s*"(?P<libname>[^"\n\.]+)"\s*$', re.MULTILINE)
        re_inc = re.compile(r'^\s*%include\s*"(?P<filename>[^"\n]+)"\s*$', re.MULTILINE)
        # Include statements without an extension are library includes (.h, plus .c if embedding the runtime)
        # McCode3 would insert the library immediately on first-encounter, but we want to collect all libraries
        # and ensure they're included before all components.

        # Get any library names in this block:
        libraries = list(dict.fromkeys([match.group('libname') for match in re_lib.finditer(raw_c)]))
        # *erase* the library includes from the block:
        raw_c = re.sub(re_lib, '', raw_c)

        # We must look through the to-be-included code *now* to see if it *also* includes more libraries.
        includes = [CInclude(parent=parent, name=lib, order=index, content=self._file_contents(f'{lib}.h'))
                    for index, lib in enumerate(libraries)]
        new_includes = []
        for include in includes:
            found_includes, include.content = self._handle_raw_c_include(include.name, include.content)
            if found_includes:
                new_includes.extend(found_includes)
                new_includes = list(dict.fromkeys(new_includes))
        if len(new_includes):
            includes.extend(new_includes)
            includes = list(dict.fromkeys(includes))

        # If there are any non-library includes remaining, we should insert them immediately:
        matches = list(re_inc.finditer(raw_c))
        count = 0
        while len(matches):
            count += 1
            # matches[0] is a re.Match object with (start, stop) = matches[0].span(0) including the full matched string
            # We can use this directly to replace the match without using re.sub, which can cause problems if there are
            # escape characters in the replacement (It seems)
            start, stop = matches[0].span(0)
            raw_c = raw_c[:start] + self._file_contents(matches[0].group('filename')) + raw_c[stop:]

            #filename = matches[0].group('filename')
            #this_re = re.compile(rf'^\s*%include\s*"{filename}"\s*$', re.MULTILINE)
            #logger.info(f'Replacing {this_re} with file contents from {filename}')
            #raw_c = re.sub(this_re, self._file_contents(filename, True), raw_c)

            # re-match since we modified raw_c
            matches = list(re_inc.finditer(raw_c))

        if count:
            found_includes, raw_c = self._handle_raw_c_include(parent, raw_c)
            if len(found_includes):
                includes.extend(found_includes)
                includes = list(dict.fromkeys(includes))

        return includes, raw_c

    def _parse_libraries_for_typedefs(self):
        from .c_listener import extract_c_defined_types as parse
        typedefs = set()
        for include in self.includes:
            # logger.debug(f'library {include.name}')
            # The files '%include'-d can themselves use the '%include' mechanism :/
            defined_types = parse(include.content, user_types=list(typedefs))
            typedefs = typedefs.union(set(defined_types))
            # logger.debug(f'{include.name} done')
        # types can also be defined in component 'SHARE' blocks:
        for block in [share for comp in self.source.component_types() if len(comp.share) for share in comp.share]:
            # TODO decide if this should be block.translated or block.source
            defined_types = parse(block.to_c(), user_types=list(typedefs))
            typedefs = typedefs.union(set(defined_types))
        self.typedefs = list(typedefs)

    def _determine_uservars(self):
        """An instrument can define 'USERVARS' which thus far has been treated as an unparsed block of code.
        We now must extract parameter declarations from this block (there should be *only* declarations, no
        initialization is supported in McCode-3).
        Each declaration *should* of the form
            {type} {name};
            {type} * {name};
            {type} {name}[N];
        Where {type} is a valid C typename _or_ a typename defined in any included libraries or 'SHARE' blocks.
        A possible future extension could allow for valid instantiation in the USERVAR blocks.
        This function extracts the parameter {name}, {type}, (scalar|pointer|array)-ness, and initializer if present
        for the unique set of all variables added to the _particle struct by the user -- it checks for name clashes
        between definitions as well (identical definitions are allowed but not equal named non-matching definitions)

        Extracted values are stored in namedtuple `CDeclaration` objects for easier interaction.
        """
        def extract_declares(name, raw_c_obj):
            # TODO decide if this should be raw_c_obj.translated or raw_c_obj.source
            c_decs = extract_c_declared_variables(raw_c_obj.to_c(), user_types=self.typedefs)
            if any(d.init is not None for d in c_decs):
                logger.critical(f'Warning USERVARS block from {name} contains assignment(s) (= sign).')
                logger.critical('        Move them to an EXTEND section. May fail at compile')
            return c_decs

        declares = []  # runtime-accessing by 'ID' only possible if order is preserved
        for raw_user_vars_c_block in self.source.user:
            # declares = declares.union(extract_declares(self.source.name, raw_user_vars_c_block))
            declares.extend(extract_declares(self.source.name, raw_user_vars_c_block))
            declares = list(dict.fromkeys(declares))

        # Check for differing repeated declarations -- these are set elements with the same name:
        if len(set([d.name for d in declares])) != len(declares):
            raise RuntimeError(f"One or more instrument USERVARS repeated in {self.source.name}")

        # Look for USERVAR section(s) in the set of component definitions too:
        comp_declares = dict()
        for component in self.source.component_types():
            # temp_set = set()
            a_declares = []
            for raw_user_vars_c_block in component.user:
                # temp_set = temp_set.union(extract_declares(component.name, raw_user_vars_c_block))
                a_declares.extend(extract_declares(component.name, raw_user_vars_c_block))
                a_declares = list(dict.fromkeys(a_declares))
            comp_declares[component.name] = a_declares
            # Check for differing repeated declarations in this component:
            if len(set([d.name for d in comp_declares[component.name]])) != len(comp_declares[component.name]):
                raise RuntimeError(f"One or more component USERVARS repeated in {component.name}")

        # Go through all components with uservars, then through the instances of that type
        # and append the instance index to the name of the uservar
        inst_declares = dict()
        for comp_name, a_declares in comp_declares.items():
            i_declares = []
            for index, instance in enumerate(self.source.components):
                if instance.type.name == comp_name:
                    i_declares.extend([d.copy(suffix=str(index+1)) for d in a_declares])
            inst_declares[comp_name] = i_declares
        # Replace the component definition 'base' uservar declares with the 'real' ones:
        comp_declares = inst_declares

        # And between all components and the instrument
        nd = set([x for dec in comp_declares.values() for x in dec]).union(declares)
        if len(set([x.name for x in nd])) != len(nd):
            culprits = [x.name for x in self.source.component_types() if len(comp_declares[x.name])]
            raise RuntimeError(f'Conflicting USERVAR declarations between instrument {self.source.name} and {culprits}')

        self.instrument_uservars = declares
        self.component_uservars = comp_declares

    def __post_init__(self):
        """Before doing *anything* else. We need to search through all raw C code for %include statements.
        This could not be done earlier if the goal is to have a clear separation between the parsing and target
        languages. (A different target language would not include the same libraries in its raw blocks)
        """
        # Make sure the registry list contains the C library registry, so that we can find and include files
        from ..reader.registry import ordered_registries
        if not any(reg == LIBC_REGISTRY for reg in self.registries):
            self.source.registries += (LIBC_REGISTRY, )
        self.source.registries = tuple(ordered_registries(list(self.source.registries)))

        includes = []
        inst = self.source
        for grp in (inst.user, inst.declare, inst.initialize, inst.save, inst.final):
            for block in grp:
                incs, block.translated = self._handle_raw_c_include(inst.name, block.source)
                includes.extend(incs)
                includes = list(dict.fromkeys(includes))
        for typ in inst.component_types():
            for grp in (typ.share, typ.user, typ.declare, typ.initialize, typ.trace, typ.save, typ.final, typ.display):
                for block in grp:
                    incs, block.translated = self._handle_raw_c_include(typ.name, block.source)
                    includes.extend(incs)
                    includes = list(dict.fromkeys(includes))
        for instance in inst.components:
            for block in instance.extend:
                incs, block.translated = self._handle_raw_c_include(instance.name, block.source)
                includes.extend(incs)
                includes = list(dict.fromkeys(includes))
        self.includes = sort_include_hierarchy(includes)

        # search the library headers for type definitions:
        self._parse_libraries_for_typedefs()

        # pull together the per-component-type defined parameters into a dictionary... since this is required
        # in multiple places :/  -- now using CDeclarator class objects
        for typ in inst.component_types():
            dp = []
            for block in typ.declare:
                # TODO decide if this should be block.translated or block.source
                dp.extend(extract_c_declared_variables(block.to_c(), user_types=self.typedefs))
                dp = list(dict.fromkeys(dp))
            self.component_declared_parameters[typ.name] = dp
        self._determine_uservars()

    def _instrument_and_component_uservars(self):
        uuv = [x for x in self.instrument_uservars]
        for comp_name, comp_uv in self.component_uservars.items():
            uuv.extend(comp_uv)
            uuv = list(dict.fromkeys(uuv))
        return uuv

    def visit_header(self):
        from .c_header import header_pre_runtime, header_post_runtime
        # Get the unique list of USERVAR declarations:
        # logger.debug(f'Instrument uservars = {self.instrument_uservars}')
        # uuv = set().union(self.instrument_uservars)
        # for x in self.component_uservars:
        #     uuv = uuv.union(x)

        uuv = self._instrument_and_component_uservars()

        is_mcstas = self.is_mcstas
        self.out(header_pre_runtime(is_mcstas, self.source, self.runtime, self.config, uuv))
        # runtime part
        if self.config.get('include_runtime'):
            self.out('#define MC_EMBEDDED_RUNTIME')
            self.configure_file('mccode-r.h.in', 'mcstas' if is_mcstas else 'mcxtrace')
            # self.embed_file('mcstas-d.h' if is_mcstas else 'mcxtrace-d.h')
            # self.embed_file("mccode-r.h")
            self.embed_file('mcstas-r.h' if is_mcstas else 'mcxtrace-r.h')
            self.embed_file("mccode-r.c")
            self.embed_file('mcstas-r.c' if is_mcstas else 'mcxtrace-r.c')
            if self.verbose:
                print(f"Compile with flags '-DUSE_NEXUS -lNeXus' to enable NeXus support")
        else:
            # This only works if the module is *not* a compressed archive
            # If it is, we would need to do some trickery to ... write out the
            self.out(f'#include "{self.include_path("mccode-r.h")}"')
            self.out(f'#include "{self.include_path("mcstas-r.h" if is_mcstas else "mcxtrace-r.h")}"')
            print(f"Dependency: mccode_antlr-r.o\nDependency: {'mcstas-r.o' if is_mcstas else 'mcxtrace-r.o'}")

        # # TODO insert includes here?
        # if len(self.includes):
        #     self.out("/* %include libraries from instrument and component definitions */")
        # for include in self.includes:
        #     self.out(f'/* Contents of {include.name}.h (requested from {include.parent})*/')
        #     self.out(include.content)

        self.out(header_post_runtime(self.source, self.runtime, self.config, self.include_path()))

    def include_header(self, header: CInclude):
        self.info(f'include {header} header')
        self.out(f'/* Contents of {header.name}.h (requested from {header.parent}) */')
        self.out(header.content)

    def include_source(self, source: CInclude):
        self.info(f'include {source} source')
        # c files can %include *more* files! If the specified files have already been included, there's nothing
        # to do. Otherwise, panic.
        lib_names = [lib.name for lib in self.includes]
        c_name = f'{source.name}.c'
        c_libraries, c_content = self._handle_raw_c_include(c_name, self._file_contents(c_name))
        for c_lib in [c_lib for c_lib in c_libraries if c_lib.name not in lib_names]:
            logger.info(f'{c_name} requested {c_lib.name} but is has not been included yet!')
            self.include_header(c_lib)
            self.include_source(c_lib)
            #self.includes.append(c_lib)
        self.out(f'/* Contents of {source.name}.c (requested from {source.parent}) */')
        self.out(c_content)

    def visit_declare(self):
        from .c_decls import declarations_pre_libraries
        self.info("Writing instrument and components DECLARE")

        if len(self.includes):
            self.out("/* %include libraries from instrument and component definitions */")
        for include in self.includes:
            self.include_header(include)

        self.info('Pre library declarations')
        contents, warnings = declarations_pre_libraries(self.source, self.typedefs, self.component_declared_parameters)
        self.out(contents)

        self.info('Include runtime / list dependencies')
        if self.config.get('include_runtime'):
            for include in self.includes:
                self.include_source(include)
        else:
            for include in self.includes:
                print(f'Dependency: {include.name}.o')

        self.out("/* User declarations from instrument definition. Can define functions. */")
        self.out('\n'.join([dec.to_c() for dec in self.source.declare]))
        # FIXME I _think_ these macro undefines are not used (that is, they're never defined in the first place)
        self.out('#undef compcurname\n#undef compcurtype\n#undef compcurindex')
        self.out(f"/* end of instrument '{self.source.name}' and components DECLARE */")

        self.warnings += warnings

    def visit_initialize(self):
        from .c_initialise import cogen_initialize
        self.out(cogen_initialize(self.source, self.component_declared_parameters, self.ok_to_skip))

    def enter_trace(self):
        from .c_trace import def_trace_section, cogen_trace_section
        self.out(def_trace_section(self.is_mcstas))
        self.out(cogen_trace_section(self.is_mcstas, self.source, self.component_declared_parameters,
                                     self.instrument_uservars, self.component_uservars))

    def leave_trace(self):
        from .c_trace import undef_trace_section
        self.out(undef_trace_section(self.is_mcstas))

    def enter_uservars(self):
        # After cogen_trace_section, the USERVAR particle struct members need to be defined for raytrace
        uuv = self._instrument_and_component_uservars()
        self.out('\n'.join([f'#define {x.name} (_particle->{x.name})' for x in uuv]))

    def leave_uservars(self):
        # following TRACE before undef_trace_section, the USERVAR particle struct members need to be undef'd
        uuv = self._instrument_and_component_uservars()
        self.out('\n'.join([f'#undef {x.name}' for x in uuv]))

    def visit_raytrace(self):
        from .c_raytrace import cogen_raytrace
        self.out(cogen_raytrace(self.source, self.ok_to_skip))

    def visit_raytrace_funnel(self):
        from .c_raytrace import cogen_funnel
        self.out(cogen_funnel(self.source, self.ok_to_skip))

    def visit_save(self):
        from .c_save import cogen_save
        self.out(cogen_save(self.source, self.component_declared_parameters))

    def visit_finally(self):
        from .c_finally import cogen_finally
        self.out(cogen_finally(self.source, self.component_declared_parameters))

    def visit_display(self):
        from .c_display import cogen_display
        self.out(cogen_display(self.source, self.component_declared_parameters))

    def visit_macros(self):
        from .c_macros import cogen_getvarpars_fct, cogen_getcompindex_fct, cogen_getparticlevar_fct
        self.out(cogen_getvarpars_fct(self.source))
        self.out(cogen_getparticlevar_fct(self._instrument_and_component_uservars()))
        self.out(cogen_getcompindex_fct(self.source))
        self.embed_file('metadata-r.c')
        self.embed_file('mccode_main.c')
        self.out(f'/* end of generated C code for {self.source.name} */')



"""The McCode nightly test script implemented for mccode-antlr."""

from dataclasses import dataclass, field
from datetime import timedelta
from pathlib import Path
from typing import TypeVar
from loguru import logger

from numpy import nan

from mccode_antlr.reader import Registry

NightlyInstrExampleType = TypeVar('NightlyInstrExampleType', bound='NightlyInstrExample')


@dataclass
class NightlyInstrExample:
    """Holds parameters for a single instr file %Example: test case"""
    sourcefile: Path
    test_number: int
    parameter_values: str = ""
    detector: str = ""
    target_value: float = nan
    test_value: float = nan
    compiled: bool = False
    compile_time: timedelta = field(default_factory=timedelta)
    ran: bool = False
    run_time: timedelta = field(default_factory=timedelta)
    error_message: str = ""
    stdout: str = ""
    test_complete: bool = False

    def __post_init__(self):
        import re
        if '.instr' in self.parameter_values:
            rinstr = re.compile(r'[a-zA-Z][a-zA-Z0-9_\-]*\.instr')
            new_param = rinstr.sub('', self.parameter_values)
            logger.info(f'Removing .instr from {self.parameter_values} now {new_param}')
            self.parameter_values = new_param
        if self.sourcefile.stem in self.parameter_values:
            new_param = self.parameter_values
            self.parameter_values = new_param.replace(self.sourcefile.stem, '')
            logger.info(f'Removing {self.sourcefile.stem} from {new_param} now {self.parameter_values}')
        if '-n' in self.parameter_values:
            rn = re.compile(r'-n[0-9eE\.]*')
            new_param = rn.sub('', self.parameter_values)
            logger.info(f'Removing -n from {self.parameter_values}, now {new_param}')
            self.parameter_values = new_param

    @classmethod
    def list_from_file(cls, filename: Path) -> list[NightlyInstrExampleType]:
        import re
        rspec = re.compile(f'%Example:([^\n]*)Detector:([^\n]*)_I=([0-9.+-e]+)')
        with open(filename, 'r') as file:
            contents = file.read()
            matches = rspec.findall(contents)
        if len(matches):
            groups = [(m[0].strip(), m[1].strip(), float(m[2].strip())) for m in matches]
            return [cls(filename, i+1, p, d, t) for i, (p, d, t) in enumerate(groups)]
        return [cls(filename, -1)]

    def get_display_name(self):
        return f'{self.sourcefile.stem}_{self.test_number}' if self.test_number > 1 else self.sourcefile.stem

    def get_json_repr(self):
        jr = dict(displayname=self.get_display_name(),
                  sourcefile=self.sourcefile.as_posix(),
                  testnb=self.test_number,
                  parvals=self.parameter_values,
                  detector=self.detector,
                  targetval=self.target_value,
                  testval=self.test_value,
                  compiled=self.compiled,
                  compiletime=str(self.compile_time),
                  didrun=self.ran,
                  runtime=str(self.run_time),
                  errmsg=self.error_message,
                  stdout=self.stdout,
                  complete=self.test_complete
                  )
        return jr

    def save(self, path: Path):
        import json
        with open(path.joinpath(self.get_display_name() + '.json'), 'w') as file:
            json.dump(self.get_json_repr(), file)


@dataclass
class LineLogger:
    lines: list = field(default_factory=list)

    def logline(self, line: str):
        self.lines.append(line)

    def save(self, filename: Path):
        with open(filename, 'w') as file:
            file.write('\n'.join(self.lines))
            file.write('\n')

    def find(self, search: str):
        import re
        for line in self.lines:
            if re.search(search, line):
                return True
        return False


def get_cpu_name():
    from platform import system
    from subprocess import check_output
    if system() == "Windows":
        from platform import processor
        return processor()
    elif system() == "Darwin":
        from os import environ, pathsep
        environ['PATH'] = environ['PATH'] + pathsep + '/usr/sbin'
        command = "sysctl -n machdep.cpu.brand_string"
        return check_output(command).strip()
    elif system() == "Linux":
        command = "cat /proc/cpuinfo"
        all_info = check_output(command, shell=True).decode().strip()
        for line in all_info.split("\n"):
            if "model name" in line:
                from re import sub
                return sub(".*model name.*:", "", line, 1).strip()
    return ""


def get_nvidia_gpu_name():
    """Identify GPU (Uses GPUtil which only knows about Nvidia GPUs)"""
    from importlib import import_module
    from importlib.util import find_spec
    if find_spec('GPUtil') is None:
        return "none"
    gpus = import_module('GPUtil').getGPUs()
    names = ','.join(gpu.name for gpu in gpus)
    return names if len(names) else "none"


def get_hostname():
    import socket
    return socket.gethostname()


def get_username():
    import os
    import pwd
    info = pwd.getpwuid(os.getuid())
    # Return the full name 'pw_gecos' if present, otherwise the username 'pw_name'
    return info[4] if len(info[4]) else info[0]


def _monitor_name_file_name_match(folder, monitor_name):
    import re
    look_for_filename = False
    with open(folder.joinpath('mccode.sim'), 'r') as file:
        lines = file.readlines()
    for line in lines:
        if re.match(rf"\s*component:\s*{monitor_name}$", line):
            look_for_filename = True
        if look_for_filename:
            m = re.match(r'\s*filename:\s+(.+)', line)
            if m:
                filename = folder.joinpath(m.group(1))
                if filename.is_file():
                    return filename
            if re.match('end data', line):
                # try the monitor named dat file
                zero_d_filename = folder.joinpath(monitor_name + '.dat')
                if zero_d_filename.is_file():
                    return zero_d_filename
                return None
    return None


def mccode_test_compiler(work_dir, file_path, target, registry, generator, dump, **kwargs):
    from pathlib import Path
    from mccode_antlr.reader import Reader
    from mccode_antlr.compiler.c import compile_instrument
    from mccode_antlr.common import Mode
    # only the provided (remote) registry should be necessary for test instruments
    reader = Reader(registries=[registry])
    inst = reader.get_instrument(file_path, mode=Mode.minimal)
    output = Path(work_dir)
    config = dict(default_main=True, enable_trace=False, portable=False, include_runtime=True,
                  embed_instrument_file=False, verbose=False)
    try:
        binary_path = compile_instrument(inst, target, output, generator=generator, config=config, dump_source=dump, **kwargs)
    except RuntimeError as err:
        logger.critical(f'Failed to produce a binary for {file_path}')
        logger.info(err)
        return False, Path()
    return True, binary_path


def mcstas_test_compiler(target, work_dir, file_path, dump, **kwargs):
    from mccode_antlr.reader import MCSTAS_REGISTRY
    from mccode_antlr.translators.target import MCSTAS_GENERATOR
    return mccode_test_compiler(work_dir, file_path, target, MCSTAS_REGISTRY, MCSTAS_GENERATOR, dump, **kwargs)


def mcxtrace_test_compiler(target, work_dir, file_path, dump, **kwargs):
    from mccode_antlr.reader import MCXTRACE_REGISTRY
    from mccode_antlr.translators.target import MCXTRACE_GENERATOR
    return mccode_test_compiler(work_dir, file_path, target, MCXTRACE_REGISTRY, MCXTRACE_GENERATOR, dump, **kwargs)


def mccode_test_runner(target, binary_path, test_parameters: str, n_particles: int):
    import shutil
    from tempfile import mkdtemp
    from mccode_antlr.compiler.c import run_compiled_instrument

    output_path = Path(mkdtemp(prefix=binary_path.stem, dir=binary_path.parent.resolve()))
    # The McCode C runtime won't use an existing directory ...
    shutil.rmtree(output_path)
    # common default parameters: output directory, seed value, number of particles
    parameters = f'--dir {output_path} -s 1000 -n {n_particles} ' + test_parameters
    try:
        output = run_compiled_instrument(binary_path, target, parameters, capture=True)
    except RuntimeError as error:
        return False, str(error), output_path
    return True, output, output_path


def mcstas_test_runner(target, binary_path, test_parameters: str, n_particles: int):
    # If NeXus output is requested and the InstrumentDescriptionFile is needed, run a different script entirely...
    #   TODO think about actually doing this?
    # if target.flags & CBinaryTarget.Type.nexus and idf_required:
    #     run([config['idfgen'].get(str), str(binary), *options])
    return mccode_test_runner(target, binary_path, test_parameters, n_particles)


def mcxtrace_test_runner(target, binary_path, test_parameters: str, n_particles: int):
    return mccode_test_runner(target, binary_path, test_parameters, n_particles)


def mccode_test(compiler, runner, mpi: int = None, acc: bool = False, nexus: bool = False, **kwargs):
    """Specialize the compiler and runner based on requested number of processors and use of OpenACC, then test"""
    from functools import partial
    from mccode_antlr.compiler.c import CBinaryTarget
    target = CBinaryTarget(mpi=mpi is not None, acc=acc, count=1 if mpi is None else mpi, nexus=nexus)
    return _mccode_test(partial(compiler, target), partial(runner, target), **kwargs)


def mcstas_test(search_pattern=None, instr_count=None, skip_non_test: bool = False,
                mpi: int = None, acc: bool = False, nexus: bool = False, n_particles: int = 1000,
                dump: bool = False, workdir: Path = None):
    from mccode_antlr.reader import MCSTAS_REGISTRY as REGISTRY
    return mccode_test(mcstas_test_compiler, mcstas_test_runner, mpi, acc, nexus, registry=REGISTRY,
                       search_pattern=search_pattern, instr_count=instr_count, skip_non_test=skip_non_test,
                       n_particles=n_particles, dump=dump, workdir=workdir)


def mcxtrace_test(search_pattern=None, instr_count=None, skip_non_test: bool = False,
                  mpi: int = None, acc: bool = False, nexus: bool = False, n_particles: int = 1000,
                  dump: bool = False, workdir: Path = None):
    from mccode_antlr.reader import MCXTRACE_REGISTRY as REGISTRY
    return mccode_test(mcxtrace_test_compiler, mcxtrace_test_runner, mpi, acc, nexus, registry=REGISTRY,
                       search_pattern=search_pattern, instr_count=instr_count, skip_non_test=skip_non_test,
                       n_particles=n_particles, dump=dump, workdir=workdir)


def _mccode_test(compiler, runner, registry: Registry, search_pattern=None, instr_count=None,
                skip_non_test: bool = False, n_particles: int = 1000, workdir: Path = None, dump: bool=False):
    """Run McCode instrument test(s) using the provided `compiler` and `runner` functions to handle those tasks.

    Each takes the expected working directory name (a string) and the path to the instr file as input.
    The compiler should return only a boolean success indication.
    The runner should return a boolean success indicator, the contents of the run standard output, and the directory
    (if any) used to store run output files -- these files are intended to reside below the working directory and will
    be automatically removed by this function.

    Parameters:
        compiler: a function taking a string and Path and producing the necessary runtime executable
        runner: a function taking a string and Path and running the executable, storing its output
        registry: a Registry object which knows where to find all 'examples/*' instr files
        search_pattern: an optional positive filter for the registered 'examples/*' instr files
        instr_count: an optional positive count of filtered instr files to include
        skip_non_test: a flag to control whether instr files which do not specify test cases should be compiled and run
    """
    import re
    from loguru import logger as logging
    import tempfile
    from datetime import datetime
    logging.info(f"Finding test instruments in: {registry}")
    # we _require_ that all test instr files are in a folder called "examples" ...
    filenames = registry.match(re.compile(r'.*examples/.*\.instr'))
    # allow the user to limit which test cases can be found via a regular expression search
    if search_pattern is not None:
        if not isinstance(search_pattern, re.Pattern):
            search_pattern = re.compile(search_pattern)
        filenames = [x for x in filenames if search_pattern.search(x) is not None]
    # the examples folder has at least one script in it:
    filenames = [x for x in filenames if Path(x).suffix == '.instr']
    # optionally test case count limiter:
    if instr_count is not None and isinstance(instr_count, int) and instr_count > 0:
        filenames = filenames[:instr_count]
    # grab full paths to each file (in case this is a remote registry, they need to be fetched)
    filepaths = [registry.path(filename) for filename in filenames]

    longest_name = max(len(x.stem) for x in filepaths)

    # Build test objects from the '%Example:' line(s) in each file:
    tests = [x for filepath in filepaths for x in NightlyInstrExample.list_from_file(filepath)]

    binaries = dict()

    # By default, use a temporary directory -- but allow the user to specify someplace else:
    created = False
    if workdir is None:
        created = True
        workdir = tempfile.mkdtemp()
    if not isinstance(workdir, Path):
        workdir = Path(workdir)
    if not workdir.exists():
        created = True
        workdir.mkdir(parents=True)
    if not workdir.is_dir():
        raise RuntimeError(f'Expected {workdir} to be a directory, but it is not.')

    # Move into a temporary directory to ensure compilation/run times are from scratch

    logging.info("Compiling instruments")
    for test in tests:
        print(test)
        if test.sourcefile not in binaries and (test.test_number != 0 or not skip_non_test):
            t1 = datetime.now()
            binaries[test.sourcefile] = compiler(workdir, test.sourcefile, dump=dump, recompile=True, replace=False)
            test.compile_time = datetime.now() - t1
            test.compiled = True

            if binaries[test.sourcefile][0]:
                logging.info(f'%-{test.sourcefile.stem:>{longest_name}s}: {test.compile_time}')
            else:
                logging.info(f'%-{test.sourcefile.stem:>{longest_name}s}: COMPILE ERROR')
        else:
            logging.info(f'%-{test.sourcefile.stem:>{longest_name}s}: COMPILE skipped')

    logging.info("Running tests")
    for test in tests:
        if not binaries.get(test.sourcefile, (False, None))[0]:
            logging.info(f'%-{test.sourcefile.stem:>{longest_name}s}: No compile')
            continue
        if test.test_number < 1:
            logging.info(f'%-{test.sourcefile.stem:>{longest_name}s}: No test')
            continue

        t1 = datetime.now()
        logger.info(f'run {binaries[test.sourcefile][1]} -n {n_particles} {test.parameter_values}')
        test.ran, stdout, output_dir = runner(binaries[test.sourcefile][1], test.parameter_values, n_particles)
        test.run_time = datetime.now() - t1
        test.stdout = stdout.decode() if isinstance(stdout, bytes) else stdout

        if test.ran:
            logging.info(f'%-{test.sourcefile.stem:>{longest_name}s}: {test.run_time}')
        else:
            logging.info(f'%-{test.sourcefile.stem:>{longest_name}s}: RUNTIME ERROR')
            logging.debug(test.stdout)
            test.error_message = test.stdout
            continue

        detector_output = _monitor_name_file_name_match(output_dir, test.detector)
        logging.debug(f'Detector output stored in {detector_output}')
        test.test_value = - 1
        if detector_output is None or not detector_output.is_file():
            rdet = re.compile(f'{test.detector}_I= ([0-9+-eE.]+)')
            matches = rdet.findall(test.stdout)
            if len(matches):
                test.test_value = float(matches[0][0])
        else:
            with open(detector_output, 'r') as detector_file:
                detector_lines = detector_file.readlines()
            detector_pattern = re.compile('# values: ([0-9+-e.]+) ([0-9+-e.]+) ([0-9]+)')
            detector_matches = [detector_pattern.match(line) for line in detector_lines if detector_pattern.match(line)]
            if len(detector_matches):
                test.test_value = float(detector_matches[0].group(1))

        test.test_complete = True

    if created:
        import shutil
        shutil.rmtree(workdir)
    else:
        logging.info(f'Test compiled binaries and test case output directories are located under {workdir}.\n'
                     'You may wish to clean-up this directory to recover disk space.')

    # Since the output directories may no longer exist, save the test results ... here?
    for test in tests:
        test.save(Path())

    results = {t.get_display_name(): t.get_json_repr() for t in tests}
    # the 'ncount' and 'mpi' keys should be defined elsewhere
    results['_meta'] = {'date': datetime.now().isoformat(), 'hostname': get_hostname(),
                        'user': get_username(), 'cpu_type': get_cpu_name(), 'gpu_type': get_nvidia_gpu_name()}
    return results



def ispath(arg: str):
    from pathlib import Path
    return Path(arg)


def main(name, program):
    from argparse import ArgumentParser
    from loguru import logger
    parser = ArgumentParser(name, description=f'Test instrument compilation and runtime for {name}')
    parser.add_argument('-s', '--search', help='Regular expression positive filter for instrument names', default=None)
    parser.add_argument('-c', '--count', type=int, help='Maximum number of instruments to test', default=None)
    parser.add_argument('-k', '--skip', action='store_true', help='Skip instruments without test cases')
    parser.add_argument('--mpi', type=int, help='Number of mpi processes to use', default=None)
    parser.add_argument('--nexus', action='store_true', help='Whether to use NeXus output')
    parser.add_argument('-n', '--ncount', type=int, help='Number of particles to simulate per test', default=1000)
    parser.add_argument('-d','--dump', action='store_true', help='Output C source to file', default=False)
    parser.add_argument('-w','--workdir', type=ispath, help='Work directory, temporary if None', default=None)

    # logging.basicConfig(level=logging.DEBUG, format='{asctime} {levelname} {message}', style='{')
    # logging.basicConfig(level=logging.DEBUG)

    args = parser.parse_args()
    results = program(search_pattern=args.search, instr_count=args.count, skip_non_test=args.skip, mpi=args.mpi,
                      nexus=args.nexus, n_particles=args.ncount, dump=args.dump, workdir=args.workdir)
    results['_meta']['mpi'] = args.mpi
    results['_meta']['ncount'] = args.ncount

def mcstas_test_main():
    return main('mcstas_test', mcstas_test)

def mcxtract_test_main():
    return main('mcxtrace_test', mcxtrace_test)


if __name__ == '__main__':
    mcstas_test_main()


from loguru import logger
"""
/*******************************************************************************
* cogen_decls: write the declaration part from the instrument description
*   that is the particle definition, the instrument parameters, the DECLARE part,
*   and all SHARE sections from components.
* input:  an instrument definition structure
* output: number of warnings/errors to fix.
*
* code is generated at root level of C file: only C definitions, no initialisers !
* calls: cogen_comp_declare
*******************************************************************************/
"""


def declarations_pre_libraries(source, typedefs: list, component_declared_parameters: dict):
    warnings = 0

    def instrument_parameters_struct():
        def _inner():
            if len(source.parameters) == 0:
                return f"char {source.name}_has_no_parameters;"
            return '\n'.join([f'  {p.value.mccode_c_type} {p.name};' for p in source.parameters])
        lines = ['struct _struct_instrument_parameters{', _inner(), '};',
                 'typedef struct _struct_instrument_parameters _class_instrument_parameters;']
        return '\n'.join(lines)

    count = len(source.groups) + sum(len(i.jump) + (0 if i.split is None else 1) for i in source.components)

    def control_statement_logic():
        #TODO FIXME Why do we care about jump counts here? The struct only contains GROUP and SPLIT counters
        # count number of groups, all jump lengths and splits
        if count == 0:
            return ""
        lines = "\n/* instrument SPLIT and GROUP control logic */\nstruct instrument_logic_struct{\n"
        comment = '/* equals index of scattering comp when in group */'
        lines += '\n'.join(f'  long Group_{name}; {comment}' for name in source.groups)
        c1 = '/* this is the SPLIT counter decremented down to 0 */'
        c2 = '/* this is the particle to duplicate */'
        lines += '\n'.join(f'  long Split_{i.name}; {c1}\n  _class_particle Split_{i.name}_particle; {c2}'
                           for i in source.components if i.split is not None)
        lines += '};\n'
        return lines

    def instrument_structure():
        n2 = len(source.components) + 1  # offset enables 1-based indexing at the cost of binary size/memory use
        lines = ["struct _instrument_struct {",
                 f"  char   _name[256]; /* the name of this instrument e.g. '{source.name}' */",
                 "/* Counters per component instance */",
                 f"  double counter_AbsorbProp[{n2}]; /* absorbed events in PROP routines */",
                 f"  double counter_N[{n2}], counter_P[{n2}], counter_P2[{n2}]; /* event counters after each component instance */",
                 f"  _class_particle _trajectory[{n2}]; /* current trajectory for STORE/RESTORE */",
                 "/* Components position table (absolute and relative coords) */",
                 f"  Coords _position_relative[{n2}]; /* positions of all components */",
                 f"  Coords _position_absolute[{n2}];",
                 "  _class_instrument_parameters _parameters; /* instrument parameters */"]
        if count:
            lines.append("struct instrument_logic_struct logic; /* instrument logic */")
        lines.append("} _instrument_var;")
        lines.append("struct _instrument_struct *instrument = & _instrument_var;")
        lines.append("#pragma acc declare create ( _instrument_var )")
        lines.append("#pragma acc declare create ( instrument )")
        return '\n'.join(lines)

    def instrument_parameters_table():
        def nps(x):
            """`None`-protected (single) double-quoted string"""
            z = '' if x is None else f'{x}'
            if z == '"':
                logger.info('single double-quote found in parameter value -- escaping to avoid C syntax error')
                z = '\"'
            return z if len(z) and z[0] == '"' and z[-1] == '"' else f'"{z}"'

        def one_line(name, typename, value, unit):
            return f'  {{"{name}", &(_instrument_var._parameters.{name}), {typename}, {nps(value)}, {nps(unit)}}},'

        lines = [f'int numipar = {len(source.parameters)};', 'struct mcinputtable_struct mcinputtable[] = {']
        lines.extend([one_line(p.name, p.value.mccode_c_type_name, p.value, p.unit) for p in source.parameters])
        lines.extend(['  {NULL, NULL, instr_type_double, "", ""}'])
        lines.append("};")
        return '\n'.join(lines)

    def metadata_table():
        def one_line(defined_by, name, mimetype, value):
            from ..common.utilities import escape_str_for_c
            return f' {{"{defined_by}", "{name}", "{mimetype}", "{escape_str_for_c(value)}"}}, '

        metadata = source.collect_metadata()
        lines = ['struct metadata_table_struct metadata_table[] = {']
        lines.extend([one_line(m.source.name, m.name, m.mimetype, m.value) for m in metadata])
        lines.extend(['  {"", "", "", ""}', '};', f'int num_metadata = {len(metadata)};'])
        return '\n'.join(lines)

    def component_share_declarations():
        components = source.component_types()  # This is a set, so inclusion order is not preserved.
        sharers = [c for c in components if len(c.share)]
        if len(sharers) == 0:
            return ''
        lines = [
            "/* ************************************************************************** */",
            "/*             SHARE user declarations for all components                     */",
            "/* ************************************************************************** */"]
        for comp in sharers:
            lines.append(f"/* Shared user declarations for all components types '{comp.name}'. */")
            # TODO FIXME to_c includes the `#line` preprocessor directive -- which was removed from McCode3??
            lines.extend([share.to_c() for share in comp.share])
        lines.extend([
            "/* ************************************************************************** */",
            "/*             End of SHARE user declarations for all components              */",
            "/* ************************************************************************** */"])
        return '\n'.join(lines)

    def component_type_declarations():
        lines = ["/* ********************** component definition declarations. **************** */"]
        lines.extend([component_type_declaration(comp, typedefs, component_declared_parameters[comp.name])
                      for comp in source.component_types()])
        return '\n'.join(lines)

    def component_instance_declarations():
        lines = [component_instance_declaration(instance, index) for index, instance in enumerate(source.components)]
        lines.append(f'int mcNUMCOMP = {len(source.components)};')
        return '\n'.join(lines)

    contents = [
        '/* *****************************************************************************',
        f"* instrument '{source.name}' and components DECLARE",
        '***************************************************************************** */',
        '',
        '/* Instrument parameters: structure and a table for the initialisation',
        '   (Used in e.g. inputparse and I/O function (e.g. detector_out) */',
        '',
        instrument_parameters_struct(),
        control_statement_logic(),
        instrument_structure(),
        instrument_parameters_table(),
        metadata_table(),
        component_share_declarations(),
        component_type_declarations(),
        component_instance_declarations(),
    ]
    return '\n'.join(contents), warnings


def component_type_declaration(comp, typedefs: list, declared_parameters: list):
    """Declare the *component type* structures needed for component instances.
    Includes the component parameters structure and positioning code.
    """
    # Note for future reference against McCode-3:
    # The implementation of parameter handling in cogen.c is very convoluted. It reads component-declared parameters
    # from the component definition DECLARE block, then *replaces* the output parameter list entirely by the found
    # 'decl_par' list (at one point it combined them, but that is commented out now).
    # The call tree for functions that access `comp->def->out_par` is such that the pointer is not used before it
    # is replaced, so at least there is no ambiguity between DECLARE-found parameters and OUTPUT PARAMETERS in cogen.

    lines = [
        f"/* Parameter definition for component type '{comp.name}' */",
        f'struct _struct_{comp.name}_parameters {{',
        f"  /* Component type '{comp.name}' setting parameters */"
    ]
    # TODO Veryify that the cogen.c iteration over `comp->def->set_par` does not somehow include DEFINITION PARAMETERS
    for par in comp.setting:
        # FIXME We CANT define a static array here without knowing the maximum size used in every component!
        # # if par.value.is_a(Value.Type.float_array) or par.value.is_a(Value.Type.int_array):
        # if par.value.vector_known:
        #     # this is only possible if the value is a tuple of numbers, so no str representations of calculations
        #     c_type = par.value.mccode_c_type.translate(str.maketrans('', '', ' *'))  # strip the trailing ' *'
        #     lines.append(f'  {c_type} {par.name}[{par.value.vector_len}];')
        # el
        if par.value.is_str:
            # TODO FIXME The cogen implementation *ALSO* makes static arrays for `vector` parameters?

            # the mccode_antlr runtime does not want to allocate or deallocate the memory for this string.
            # So it expects a statically sized (16kB) char array
            c_type = par.value.mccode_c_type.translate(str.maketrans('', '', ' *'))
            lines.append(f'  {c_type} {par.name}[16384];')
        else:
            # basic integer or float
            lines.append(f'  {par.value.mccode_c_type} {par.name};')

    # This is the loop over the *replaced* `comp->def->out_par` e.g., found DECLARE parameters
    lines.append(f"  /* Component type '{comp.name}' private parameters */")
    lines.extend([f'  {x.as_struct_member()};' for x in declared_parameters])

    if len(comp.setting) + len(declared_parameters) == 0:
        lines.append(f'  char {comp.name}_has_no_parameters;')

    lines.extend([
        f'}}; /* _struct_{comp.name}_parameters */',
        f'typedef struct _struct_{comp.name}_parameters _class_{comp.name}_parameters;',
        '',
        f"/* Parameters for component type '{comp.name}' */",
        f"struct _struct_{comp.name} {{",
        # TODO We *could* look through all instances of comp to find the longest name
        f"  char     _name[256]; /* e.g. instance of {comp.name} name */",
        f"  char     _type[{len(comp.name)+1}]; /* {comp.name} */",
        f"  long     _index; /* index in TRACE list */",
        "  Coords   _position_absolute;",
        "  Coords   _position_relative; /* wrt PREVIOUS */",
        "  Rotation _rotation_absolute;",
        "  Rotation _rotation_relative; /* wrt PREVIOUS */",
        "  int      _rotation_is_identity;",
        "  int      _position_relative_is_zero;",
        f"  _class_{comp.name}_parameters _parameters;",
        "};",
        f"typedef struct _struct_{comp.name} _class_{comp.name};"
    ])
    return '\n'.join(lines)


def component_instance_declaration(instance, index):
    """Declare the *instance* variable of the component-type structure"""
    lines = [
        f'/* component {instance.name} = {instance.type.name}() [{1 + index}] DECLARE */',
        f'_class_{instance.type.name} _{instance.name}_var;',
        f'#pragma acc declare create ( _{instance.name}_var )'
    ]
    return '\n'.join(lines)
